#!/usr/bin/env python
# coding: utf-8

import numpy as np

from scipy.spatial import KDTree
from scipy.spatial.distance import cdist

import collections # for defaultdict, namedtuple

from . import IS_stat
from .candynodes import CandyNodes
from . import sball 


        
        

#č vlastní slovník se dycky hodí
class WardrobeDict(dict):
    def __init__(self):
        super().__init__()
        self.index = 0
        
    def add(self, item):
        #č nulový index se nesmí vratit
        #č CV ho použivá jako False
        self.index += 1
        #оӵ мар ке вал отын - туж жаль
        self[self.index] = item
        return self.index



TData = collections.namedtuple('TData', ('X', 'imask', 'dd'))
#VorSTM = collections.namedtuple('VorSTM', \
#       ('nsim', 'failure', 'mixed', 'pfv', 'pfw'))

## TRI-compatible estimation
#TriSTM = collections.namedtuple('TriSTM', \
#       ('outside', 'success', 'failure', 'mixed', 'pfv', 'pfw'))

        

# HalfspaceVoronoi
# TwoFaces :)
# TwoHalvesVoronoi
# TwoHalves
class TwoHorses:
    """
    č Motivace třídy:
    č Bylo mi lito vyhazovaných informací, které normálně
    č získávám z KDTree (dd, ii) a tak hledal jsem cestu 
    č jak ty informace, ta data vůbec nevytvařit.
    č Já je skutečně nepotřebuji ve fázi vzorkování.
    č Všechno co potřebujeme - 
    č vědět zda i,j jsou nejblížší, nebo ne.
    
    č Úvod:
    č Ve výsokých dimenzích (od 15, možná až od 20) 
    č věda nezná žádný způsob najít nejblížší vzorek
    č rychlejc jak brutforsem. 
    č To je jako najit vzdálenosti do všech vzorků
    č a pak si zvolit s tou (těmi) nejmenšími.
    č KDTree ve scipy uvnitř zřejmě přechází na brutforce,
    č neboť má stejný výkon jako cdist 
    č Složitost úlohy tedy uměrna npoints*nnodes
    
    č Hlavní myšlenka třídy:
    č Využit (zase) geometrických informací.
    č Kontaktní hyperrovina mezí "i" a "j" dělí
    č prostor na dva poloprostory.
    č Stejně jako dělí vzorky i bodíky do dvou skupin.
    č Skalarním součínem jednoduše zjistíme
    č do jaké poloviny patří.
    č Bodíky patří do oblasti (i,j) jen pokud mají 
    č vzorek "i" jako nejblížší v jednom z poloprostoru
    č a "j" - v druhém.
    č Bodíky, třeba, z poloprostoru "i" nejdřív 
    č proženeme KDTree přes vzorky na straně "i" 
    č a jenom ty, které mají "i" jako nejblížší
    č pustíme dál přes KD strom na straně "j".
    č Můžeme očekávát zrychlení díky tomu, 
    č že většína vzorků mimo kontaktní oblast
    č bude vyfiltrována jíž na prvním stromě.
    č Stromy ale nejsou balancovány, jeden může být 
    č představen jedním vzorkem, druhý - zbytkem.
    č
    č Musíme být opatrní, protože kontakt nemusí
    č vůbec existovat a je třeba ujistit se, že
    č nejblížší vzorek z druhého poloprostoru
    č je globálně nejblížší. (v tom prvním musíme
    č dělat dotáz na dva nějblížších)
    
    č Neplatí:
    č Neplatí, že by stačílo v každem ze stromu 
    č zkontolovat, že příslušný vzorek (i, nebo j)
    č je nejblížší
    
    č Už chapu, že hyperroviny, normály, meze atd.
    č nehrajou až takovou roli. Stačilo by 
    č i náhodně rozdělit vzorky na dvě sady tak,
    č aby v jedné sadě bylo i-čko, v druhé - j-čko.
    
    č Za úkol máme jednodušší úkol: vyhodit bodíky, 
    č které nejsou nejblížší
    č Možné variánty strategie (od blbého k chytejšímu):
    č 1. Smyčka. Když někdo jinej je blížší, tak s bodíkem
    č    se loučíme. Cečkovej kód knihoven ale nedoženeme...
    #for vzorečíček in vzorky: if dist(vzočíček) < l_i: ...
    č 2. Nasekat vzorky na 100500 skupin (nsim >> 100500)
    č    V každé skupině bereme dva nejblížších a vyhazujeme
    č    bodíky, pokud "i", nebo "j" nejsou nejblížší. 
    č    Na konci výber globálních sousedů ze všech skupin.
    č 3. Libovolně dělíme vzorky na dvě skupiny,
    č    do jedné patří i, do druhé j.
    č    a. Prohnat bodíky i-tou skupinou,
    č       vyhodit co není k "i" nejblížší,
    č       vzít vzdálenosti do dvou nejblížších.
    č    b. Opakovat pro j-tou, zas dva nejblížších
    č    c. Hledáme globální min ze čtyrž vzorků
    č 4. Dělíme vzorky na dvě skupíny libovolně
    č    proloženou hyperrovinou mezí nima.
    č    Můžeme na dvě skupiny dělit i bodíky.
    č    Moc to nepřínáší, jenom že budou blíž
    č    k svým globálním sousedům a efektivnějc maskovány.
    č             | vzorky "j" : efektivně jedna z ndim
    č    bodíky i |            : dimenzí ztracena,
    č                            rozhodování je podle 
    č                             zbývajících ndim-1
    č             | vzorky "j" : provedena klasterizace
    č             | bodíky j   : podel jedné dimenze
    č                            ekvivalentní jednomu kroku KDTree
    č 5. Dělíme vzorky na dvě skupíny kontaktní (kolmou)
    č    hyperrovinou přesně v půlce mezí "i" a "j":
    č    hyperrovinou je garantovano, že patří-li bodík 
    č    z "i" do kontaktu (i,j), tak je nejblížší. 
    č    Zbyde zkontrolovat ten druhý.
    č    a. Prohnat bodíky i-tou skupinou,
    č       vyhodit co není k "i" nejblížší,
    č       vzít vzdálenosti do dvou nejblížších.
    č    b. Opakovat pro j-tou, stačí jeden nejblížší
    č    c. Zkontrolujeme jenom, že d2_i > d1_j
    
    
    """
    def __init__(self, points, couple_indices):
        i, j = couple_indices
        #č vektor zadáme přímkou mezí vzorky
        #č od "i" k "j", normalizovat nebudeme
        self.normal_vector = points[j] - points[i]
        
        #č jako indice zkusme použit bod na usečce uprostřed mezí vzorky
        half_point = np.mean(points[[i, j]], axis=0) 
        
        self.delimit = np.inner(half_point, self.normal_vector)
        
        #č pokud skalarní součin je menší jak delimit
        #č (protože směr normály od "i" k "j"),
        #č tak patří hledané do poloprostoru "i"
        mask = (points @ self.normal_vector) < self.delimit 
        
        self.tree_i = KDTree(points[mask])
        self.tree_j = KDTree(points[~mask])
        
        #č neznám convinient cestu pro získaní
        #č tamtamtoho indexu na maskovaném poli
        #č np.ma mně příjde zbytečným
        # ha! monkey patching
        self.tree_i.idx = len(mask[:i][mask[:i]])
        self.tree_j.idx = len(mask[:j][~mask[:j]])
        
        
    def query(self, nodes, p_norm=2):
        #č pokud skalarní součin je menší jak delimit
        #č (protože směr normály od "i" k "j"),
        #č tak patří hledané do poloprostoru "i"
        halfmask = (nodes @ self.normal_vector) < self.delimit 
        
        nodes_i = nodes[halfmask]
        nodes_j = nodes[~halfmask]
        
        #č globální maska
        gmask = np.ones(len(nodes), dtype=bool)
        
        mask_i, dd_i = self._run(nodes_i, self.tree_i, self.tree_j, p_norm)
        mask_j, dd_j = self._run(nodes_j, self.tree_j, self.tree_i, p_norm)
        
        gmask[halfmask] = mask_i
        gmask[~halfmask] = mask_j
        
        
        dd_mask = halfmask[gmask]
        dd = np.empty((len(dd_i)+len(dd_j), 2))
        dd[dd_mask] = dd_i
        dd[~dd_mask] = dd_j
        
        return gmask, TData(X=nodes[gmask], imask=dd_mask, dd=dd)
        
        
        
    @staticmethod
    def _run(nodes, tree_1, tree_2, p_norm):
        dd, ii = tree_1.query(nodes, k=2, p=p_norm)
        mask = (ii[:, 0] == tree_1.idx)
        
        d, j = tree_2.query(nodes[mask], k=1, p=p_norm)
        mask_2 = (j == tree_2.idx)
        mask[mask] = mask_2
        
        mask_3 = dd[mask][:,1] > d[mask_2]
        mask[mask] = mask_3
        
        #č smí to být jen takto: nejdřív osa, pak maska
        #č opačné pořadí nefunguje
        dd[:,1][mask] = d[mask_2][mask_3]
        return mask, dd[mask]
        
        
        


class TwoPoints:
    def __init__(self, points, couple_indices, workers=1):
        self.couple_indices = couple_indices
        self.workers = workers
        i, j = couple_indices
        
        self.point_i = point_i = points[i]
        self.point_j = point_j = points[j]
        #č vektor zadáme přímkou mezí vzorky
        #č od "i" k "j", normalizovat nebudeme
        self.normal_vector = point_j - point_i
        
        #č jako indice zkusme použit bod na usečce uprostřed mezí vzorky
        self.half_point = np.mean(points[[i, j]], axis=0) 
        
        self.delimit = np.inner(self.half_point, self.normal_vector)
        
        #č pokud skalarní součin je menší jak delimit
        #č (protože směr normály od "i" k "j"),
        #č tak patří hledané do poloprostoru "i"
        mask = (points @ self.normal_vector) < self.delimit 
        
        #č vypnu optimizace KD stromu. 
        #č Netestoval však, zda je to skutečně rychlejší.
        self.tree_i = KDTree(points[mask], compact_nodes=False, balanced_tree=False)
        self.tree_j = KDTree(points[~mask], compact_nodes=False, balanced_tree=False)
        #č neznám convinient cestu pro získaní
        #č tamtamtoho indexu na maskovaném poli
        #č np.ma mně příjde zbytečným
        # ha! monkey patching
        self.tree_i.idx = len(mask[:i][mask[:i]])
        self.tree_j.idx = len(mask[:j][~mask[:j]])
        
        #č bacha na indexy. Musíme vyhodit i-tý a j-tý vzorky
        mask[i] = False
        self.points_i = points[mask]
        #č teď ale jedeme se stejnou, ale invertovanou maskou 
        mask[i] = True
        mask[j] = True
        self.points_j = points[~mask]
        
        
    def query(self, nodes, order='F'):
        #č uživatel získával hodně nekonzistentní výstup
        #č když posílal pouze jeden bodík.
        nodes = np.atleast_2d(nodes)
        #č pokud skalarní součin je menší jak delimit
        #č (protože směr normály od "i" k "j"),
        #č tak patří hledané do poloprostoru "i"
        halfmask = (nodes @ self.normal_vector) < self.delimit 
        
        nodes_i = nodes[halfmask]
        nodes_j = nodes[~halfmask]
        
        
        status_i = status_j = False
        if nodes_i.size:
            status_i, mask_i, dd_i = self._run(self.point_j, self.points_i, 
                                                self.tree_j, nodes_i, order)
        
        if nodes_j.size:
            status_j, mask_j, dd_j = self._run(self.point_i, self.points_j, 
                                                self.tree_i, nodes_j, order)
        
        if not (status_i or status_j):
            return False, None, None
        
        #č globální maska
        gmask = np.empty(len(nodes), dtype=bool)
        
        if status_i and not status_j:
            gmask[halfmask] = mask_i
            gmask[~halfmask] = False
            return True, gmask, TData(X=nodes[gmask], imask=mask_i[mask_i], dd=dd_i)
            
        if status_j and not status_i:
            gmask[halfmask] = False
            gmask[~halfmask] = mask_j
            return True, gmask, TData(X=nodes[gmask], imask=~mask_j[mask_j], dd=dd_j)
            
        
        gmask[halfmask] = mask_i
        gmask[~halfmask] = mask_j
        
        dd_mask = halfmask[gmask]
        dd = np.empty(len(dd_i) + len(dd_j))
        dd[dd_mask] = dd_i
        dd[~dd_mask] = dd_j
        
        return True, gmask, TData(X=nodes[gmask], imask=dd_mask, dd=dd)
    
    
    @staticmethod
    def _run(center_2, points_1, tree_2, nodes_1, order):
        """
        č Úkolem první fáze je vyhodit co nejvíc bodíků
        č Bodíky bereme z i-tého poloprostoru a kontrolní vzdálenosti - 
        č vůči vzorku "j" z protílehlé strany.
        č Předpokládáme, že na konci první fázi získame už
        č dobře vyfiltrované tečičky, které přece pro formu
        č musíme zkontrolovat i vůči vzorkům na opačné straně.
        č Předpokladáme, že zůstaly jíž vícemené dobré bodíky
        č a že smyčka druhé fáze už by musela běžet přes skoro všechno.
        č Proto se nebudeme snažit dohnat Cečkovej kód v Pythonu
        č a poslední část pokorně svěříme KDTree - 
        č nejrychlejší implementaci the nearest search na dívokém západě.
        """
        #č cdist vysloveně chce 2D pola, viděl jsem to i v kódě.
        #č Vrací sloupcovou matici. Co kdyby s F pořadí by to jelo rychlejc?
        #d1 = np.empty((len(nodes), 1), dtype=float, order=order)
        #cdist(nodes, [center_1], 'sqeuclidean', out=d1)
        d2 = np.empty((len(nodes_1), 1), dtype=float, order=order)
        cdist(nodes_1, [center_2], 'sqeuclidean', out=d2)
        
        mask = np.isfinite(d2)
        
#       #č udělám to takhle i když neumím si představit, 
#       #č jak se můžou objevit NaNs a nekonečna ve d2, když nejsou v d1
#       mask = np.isfinite(d1) & np.isfinite(d2)
        
        n = len(mask[mask])
        d_i = np.empty((n, 1), dtype=float, order=order)
        
        #č pro nás spíše budou kritické  poslední přídané vzorky
        for point in reversed(points_1):
            cdist(nodes_1[mask[:,0]], np.atleast_2d(point), 'sqeuclidean', out=d_i)
            
            mask[mask[:,0]] &= d2[mask[:,0]] <= d_i
            
            n = len(mask[mask])
            if n == 0:
                return False, None, None #mask[:,0], d_2[mask_2]
            d_i.resize((n, 1), refcheck=False)
        
        #č ty špinavé 2D matice zhora jíž nepotřebujem
        #č převedeme (co nejlevnějc) masku na 1D
        mask = mask[:,0]
        d_2, j = tree_2.query(nodes_1[mask], k=1, p=2, workers=self.workers)
        mask_2 = (j == tree_2.idx)
        
        if np.any(mask_2):
            #č hotovo. Vzdálenosti jíž nemusíme kontrolovat.
            #č ve smyčce d2 jsme jíž prohnali.
            mask[mask] = mask_2
            return True, mask, d_2[mask_2]
        else:
            return False, None, None
    
            
            
            
    
        


#оӵ кык точкаен Вороной
# KickTouchCayenneVoronoi

class ContactVoronoi:
    """
    č hlavní myšlenka-pointa třídy je, 
    č že místo integrování buňka po buňce 
    č a pofiderního rozhodování 
    č kdy a jak odhady aktualizovat
    č integrujeme zde všechny možné pary vzorků,
    č kontakt po kontaktu,
    č tj. oblastí ve kterých nejblížšimi jsou pár určitých vzorků.
    č Vyhody:
    č + Získáváme strukturální informace o problému.
    č + Nemusíme integrovat všechno. Na zelený-zelený kontakty kašleme.
    č + V rámci kontaktu entropie je neměnná, potenciál taky.
    č + Důraz integrace je právě na oblastéch zájmu, ne bůhví na čem u buňek.
    č + Nejsou-li body příslyšné kontaktu, tak řekneme, že kontakt neexistuje
    č   a vyhodíme ho z dalšího usuzování. U buňek člověk nikdy neví...
    č 
    č Nevyhody: 
    č - Kvadrát. Kontaktů je nsim-krát víc jak buňek. 
    č Ale i tak počet je konečný. Zvladnutelný. A nezavisí na dimenzi prostoru.
    
    č Další předpoklady, na kterých tuhle třídu stavíme (nevím, zda opravdu platí):
    č 1. existuje-li kontakt mezí vzorky "a" a "b", 
    č    tak aspoň nějaká část musí být obsažena v kouli mezi "a" a "b"
    č 1a. Nepatří-li střed usečky mezí vzorky "a" a "b" do kontaktu, 
    č     tak už nikdy ani nebude
    č 1b. I když nepatří střed usečky mezí vzorky "a" a "b" do kontaktu, 
    č     kontakt pořad může existovat.
    č 1c. Jde-li kontakt do nekonečna, pořad může (i když nemusí) 
    č     existovat část, která patří do (QHull) konvexní obálky.
    č 2. Při přidani vzorku "d" musíme aktualizovat všechny vazby všech 
    č    sousedících s tím d-čkem vzorků. To je právě proto, že oblast kontaktu
    č    jsou DVA sousedících vzorků. I když vzorek "d" nemá kontakt s "c",
    č    pouze s "a" a "b", pořád můžou existovat bodíky "ac" a "bc", pro které
    č    "c" a "d" stanou nejbližší.
    č 3. Už jsme opustili svět napočítaných stěn, 
    č    kde i konvexní obálku jsme vždycky sestavovali v prostoru triangulace.
    č    Konvexní obálka se může sestavovat v jiném prostoru a tak úplně
    č    každý bodík (i na usečce mezi dvěma vzorky) se může objevit venku.
    č 4a. Aktualizace. Jsou-li změny v kontaktní oblasti, tak jsou vyvolany
    č     přídáním nových vzorků (jasný) a vyfiltrováné vzorky je budou mít v ii.
    č 4b. Je-li oblast kontaktu ovlivněna přídanými vzorky, tak u vyfiltrovaných
    č     bodíků se nutné změní ii2: z předchozího ii2 na ii1, 
    č     nebo rovnou na index nového vzorku. Tj. stačí kontrolovat jenom ii2.
    č 4c. Aktualizaci lze provadět na redukovaném stromu 2 + nové vzorky 
    č
    č Neplatí:
    č 2. (Při přidani vzorku "c"). Jakože neexistují-li kontakty mezi vzorky "c" a "a",
    č    a mezi "c" a "b", tak přídaní "c" nemůže kontakt mezi "a" a "b" ovlivnit.
    
    
    č Poznámky k adaptivnímu vzorkování:
    č Pro vzorkování alike potřebujeme CoV, 
    č kde važení hustotámi je relativné,
    č násobení dvakrát aweights nezmění výslednou kovarianční matici.
    č Pro rozhodování, která sada nejlepé zachytila "buňku" - 
    č něco jako matici setrvačnosti, u které by násobění "hmotností"
    č konstantou zvětšovalo by celkovou setrvačnost systému.
    č Tušíme, že (obecně) dobře zachycená "buňka" by měla výkazovat
    č největší setrvačnost, ale i taky prostě hmotnost (pravděpodobnostní obsah).
    č Mohli bychom zůstat u CoV pro účely adaptivního vzorkování
    č a volít nejlepší sadu podle podílu pravděpodobnostního obsahu 
    č k rozptylu estimatoru 
    
    
    
    č co máme zadavat jako aweight u np.cov?
    č pdf bodíku? Nebo jejich vahy?
    č Vahy.
    č Rozepíšu zde dětailněji, protože to není na první pohled zřejmný 
    č a myslím si, že i kontrintuitivní
    č 1. Uvaha číslo jedna: když se snažíme odhadnout rozptyl vzorků,
    č    výbraných z normálního rozdělení, tak jich hustotami nepřevažíme
    č 2. Uvaha číslo dva: máme sadu IS a chceme odhadnout rozptyl skutečného
         rozdělení. Chce se tu použit váhy, že ano?
    č 3. Uvaha číslo tří: Jak bychom počítali těžiště a momenty setrvačnosti
    č    rovinného obrazce s odlíšnou tloušťkou t? 
    č    (označme "m" jako mass, v naši "fyzice" nechť bude y*t=m)
         
                sum[y_i * A_i * t_i]     Int y*t dA
         y_c = ---------------------- =  ----------
                    sum(A_i*t_i)          Int t dA

                sum[y_i * m_i]     Int y dm
         y_c = ---------------- =  ----------
                    sum(m_i)        Int  dm
         
         I = sum[(y_i - y_c)**2 * A_i * t_i] = Int (y-y_c)**2 * t dA
         
         I = sum[(y_i - y_c)**2 * m_i] = Int (y-y_c)**2 dm
         
         f_i je jako tloušťka t
         1/h_i je jako integrovační ploška dA
         w = f_i / h_i vahy jsou jako hmotnosti
         Monte Carlo - integrovaní dělením na kusy stejné hmotnosti
         IS - kusy mají odlíšnou hmotnost
         
                  
    č base_r dává poloměr, do kterého/za který spadá jistý podíl bodíků
    '
    r_ball/r_base = sigma
    '
    č Když máme nějakou oblast s poloměrem alespoň r
    č můžeme nastavit směrodatnou odchylku normálního rozdělení tak,
    č aby jisté procento bodíků dopadalo dovnitř
    č (integrujeme-li obyčejnou Voroneho buňku, tak r = mindist / 2)
    č Co ale máme dělat, když máme výběrovou směrodatnou odchylku 
    č bodíků z cenzurovanného rozdělení? Survival bias.
    č Můžu spočítat r_ball, ten nebude korrettní, protože vzorky
    č nebyly z normálního rozdělení, ale z cenzurovanného
    č A ani ten r_ball nepotřebuji. Chcu nastavit vzorkovací sigmu.
    č Co jako? Mám dělat výzkum? Počítat třetí a čtverté výběrové momenty?
    č Před tím se dělalo tohle: sigmas = sigmy / base_r
    č Což, pokud rozumím, ve 2D ještě víc změnšovalo směrodatné odchylky
    '
    č Mám výberovou "sigmu" (std) z cenzurovanného rozdělení, 
    č podle mé osy má odříznuté chvosty. 
    č Odříznuté chvosty podel jinejch os mě nezajimají
    č std * r_base(ndim) = r_ball(ndim)
    č sigma * r_base(ndim-1) = r_ball(ndim-1)
    č Nabízím opravu: sigma = std * r_base(ndim) / r_base(ndim-1)
    č Ve výsokých dimenzích rozdíl je minimální
    č Ale kdo tuší, co a jak tam funguje a má fungovat?
          
      
    č Score počítáme jako pravděpodobnostní obsah buňky dělený
    č směrodatnou odchylkou odhadu (3.) Není to ale jediná možnost
    č 1. sum((w_i - w_mean)**2) / n         To je rozptyl w
    č 2. sum((w_i - w_mean)**2) / n**2      To je rozptyl odhadu w_mean
    č 3. sqrt(sum((w_i - w_mean)**2)) / n   Směrodatná odchylka odhadu w_mean
    """
    def __init__(self, sample_box, hull, model_space='G', \
                ns=1000, p_base=0.5, auto_update=True, workers=1,\
                on_add_mixed=None, on_update_mixed=None, on_delete_mixed=None):
        self.sample_box = sample_box
        self.hull = hull
        self.model_space = model_space
        self.ns = ns
        
        self.auto_update = auto_update
        self.workers = workers
        
        if on_add_mixed is not None:
            self.on_add_mixed = on_add_mixed
        if on_update_mixed is not None:
            self.on_update_mixed = on_update_mixed
        if on_delete_mixed is not None:
            self.on_delete_mixed = on_delete_mixed
        
        #č chcu, aby behem dalších iterací pulka (p_base) tečiček dopadala do buňky
        self.base_r = sball.get_Radius_ps(sample_box.nvar, p_base)
        #č pro funkci IS_stat.sample_like(), víz. diskuse v poznamkách
        self.d = (sball.get_Radius_ps(sample_box.nvar - 1, p_base) / self.base_r)**2
        self.ndim = sample_box.nvar
        
        self._nsim = 0
        self._enabled = False # if nsim < 2 and not failsi
        self.nodes = WardrobeDict()
        # checklist
        #č defaultdict zajistí, že na nás dycky čeka seznam, 
        #č do kterého můžeme přidávat indexy bodíků
        self.couples = collections.defaultdict(list)
        self.mixed_couples = {}
        self.red_couples = {}
        #č uvnitř třídy ještě můžeme ordnung dodržovat
        self._indices_to_update = set() # set of couples to update
        self._valid_outsides = set() # node indexes
        
        
        #č já chcu, aby třída byla blbuvzdorná
        #č aby šlo ji vytvořit na prazdné skřiňce 
        #č aby neotravovala chybáma
        #self._update()
        
    
    # dump methods
    # callbacks are set up at __init__ if provided
    # otherwise do nothing
    @staticmethod
    def on_add_mixed(nodes): pass
    
    @staticmethod
    def on_update_mixed(idx): pass
    
    @staticmethod
    def on_delete_mixed(idx): pass
    
    
    """
    ## checklist
    self.couples # dict of active contacts {couple_indices: list of node idxs}
    self.mixed_couples = {}  ##č pár slovníků reprezentativních sad vzorků 
    self.red_couples = {}    ## {couple_indices: nodes}
    self._add_indices_to_update(j) 
    self._valid_outsides = set() #č zabyvá se tím bezprostředně integrovačka
    
    
#    Nodes pipeline:
#    0. Sampling: generate coordinates
#    1. _store_masked(): Filter out, create CandyNodes 
#       with d2 and (optionally) imask assigned
#    2. assign weights (w) such as sum(w)=1 over entire domain
#    3. score(): assign so called "score"
    """
        
    
    #č funkce se nemá dělat toho víc,
    #č pokud by ji někdo opakovaně volal pořad dokola
    def _update(self):
        if self.auto_update:
            if self.basic_checks():
                ##č můžeme udělat obraceně - 
                ##č je-li to nutné, nechť si uživatel třídy 
                ##č zajístí správnost autsajdů
                ##č a teprvé poté volá update()
                #č Ale nebudeme. Po aktualizaci něco i zmízí,
                #č tak si ušetříme kus práce
                self.update_contacts()
                self.invalidate_outside()
        
    
    def basic_checks(self):
        if self._enabled:
            return True
        elif self.sample_box.nsim < 2:
            return False
        elif not np.any(self.sample_box.failsi):
            return False
        else:
            self._enabled = True
            return True
            
    def invalidate_outside(self):
        self._valid_outsides.clear()
        
    #č funkce se nemá dělat toho víc,
    #č pokud by ji někdo opakovaně volal pořad dokola
    def update_contacts(self):
        #č já vím, že sample box pokážde failsi přepočítavá
        self.failsi = failsi = self.sample_box.failsi 
        self.points = points = getattr(self.sample_box, self.model_space)
        self.PDF = self.sample_box.pdf(self.model_space)
        
        assert len(self._indices_to_update) == 0
        
        nis = self.ns
        workers = self.workers
        
        #č zde postupně v cyklu prochazíme všemi (novými) páry kontaktů
        #č a omezujeme se pouse nejbližšími vzorky
        #č disjunktnost je pořad zajištěna
        #č a môžeme všechny nasbírané pravděpodobnosti jednoduše sčítat
        # good old event ids...
        # -1 = 'outside', 0=success, 1=failure, 2=mix
        for i in range(self._nsim, self.sample_box.nsim): 
            if failsi[i]: #č první je červený
                for j in range(i):
                    tp = TwoPoints(points, (i, j), workers=workers)
                    if failsi[j]:
                        #č červený kontakt
                        # -1 = 'outside', 0=success, 1=failure, 2=mix
                        self.onboard_couple(tp, event_id=1, nis=nis)
                    else:
                        #č žlutý kontakt
                        # -1 = 'outside', 0=success, 1=failure, 2=mix
                        self.onboard_couple(tp, event_id=2, nis=nis)
            
            else: #č první je zelený
                for j in range(i):
                    if failsi[j]:
                        #č žlutý kontakt
                        # -1 = 'outside', 0=success, 1=failure, 2=mix
                        tp = TwoPoints(points, (i, j), workers=workers)
                        self.onboard_couple(tp, event_id=2, nis=nis)
                    
                    #else: #č jinak nič
                        #оӵ Вож. Кулэ ӧвӧл
        
        
        #č project páry na update
        for couple_indices in self._indices_to_update:
            #č nepodporuje slovník množinové operace
            self.update_couple(couple_indices)
#            if couple_indices in self.couples:
#                #č i kdyby někdo (nelegálně!) přímo volal update_couple()
#                #č žádná katastrofa se dít nebude
#                #č defaultdict vytvoří prazdný seznam,
#                #č ale bez bodíků se aktualizace brzy skončí :)
#                self.update_couple(couple_indices)
        self._indices_to_update.clear()
        self._nsim = nsim
    
    
    def _add_indices_to_update(self, j):
        #č i-té indexy jsou ty čerstvé, přes ně iterujeme
        #č j-té - ty bežné, protí ním rozhodujeme
        if j < self._nsim:
            for k in range(j):
                self._indices_to_update.add((j, k))
            for k in range(j+1, self._nsim):
                self._indices_to_update.add((k, j))
    
    
    def _recommend_to_integration(self, nodes):
        # -1 = 'outside', 0=success, 1=failure, 2=mix
        if nodes.event_id == 2:
            self.mixed_couples[nodes.couple] = nodes
        else:
            self.red_couples[nodes.couple] = nodes
    
        
                
                
                
        
    
    def update_couple(self, couple_indices):
        #č při aktualizaci páry kontrolujeme, 
        #č zda nejsou změny u sady, doporučenou k integraci
        
        #č co všechno se tu může dít?
        #č -1. pár není, neboť smyčka tam přídavá všechny možné kombinace.
        if couple_indices not in self.couples:
            assert couple_indices not in self.mixed_couples
            assert couple_indices not in self.red_couples
            return -1
        idx_list = self.couples[couple_indices]
        
        #č -2. seznam může být prazdný i když by neměl
        if not idx_list:
            self.couples.pop(couple_indices)
            print("ContactVoronoi: empty list found for ", couple_indices)
            print(self.mixed_couples.pop(couple_indices, None))
            print(self.red_couples.pop(couple_indices, None))
            return -2
        
        #č 3. Pokud je pár červený, stačí zkontrolovat jen reprezentativní sadu
        if couple_indices in self.red_couples:
            idx = self.red_couples[couple_indices].idx
            status, nodes = self._check_nodes(idx)
            if status == 1: #č nic se nezměnilo
                return 3
            #č idx už jsme zkontrolovali. 
            idx_list.remove(idx)
            #č Dále seznam proženeme smyčkou _walk_through_couple_nodes()
            #č (i kdyby se tam poslal prazdný idx_list - nic se neděje)
            #č pokud bodíky se jen maliňko změnily (status 2 z _check_nodes), 
            #č tak budou použity jako best_nodes a dostanou se zpatky do seznamu.
            #č Propadla-li sada kompletně (status 3), tak nodes budou None
        else:
            nodes = None
        
        #č 4. (u smíšených) může být potřeba zkontorlovat všechy sady bodíků
        #č kvůli garancim skříňce
        new_idxs, best_nodes = self._walk_through_couple_nodes(idx_list, nodes)
        if not new_idxs:
            self._invalidate_couple(couple_indices)
            return -5
        else:
            self.couples[couple_indices] = new_idxs
            self._recommend_to_integration(best_nodes)
            return 4
        
        
        
    def _walk_through_couple_nodes(self, idx_list, best_nodes=None):
        new_idxs = []
        #č i kdyby se poslal prazdný idx_list - nic se neděje
        if best_nodes is not None:
            new_idxs.append(best_nodes.idx)
            best_score = best_nodes.score
        else:
            best_score = -np.inf
            
        
        sampling_needed = False
        
        for idx in idx_list:
            status, nodes = self._check_nodes(idx)
            if status != 1: #č 1 == nic se nezměnilo
                sampling_needed = True
            if status == 3: #č Padla celá sada bodů.
                continue
                
            new_idxs.append(idx)
            if nodes.score > best_score:
                best_nodes = nodes
                best_score = nodes.score
                
        if (best_nodes is not None) and sampling_needed:
            #č tp-čko se musí znovu vygenerovat na aktuálních vzorcích
            #č nepůjde je ukladat
            tp = TwoPoints(self.points, best_nodes.couple, workers=self.workers)
            nodes = self.sample_alike(tp, best_nodes, self.ns)
            if nodes is not None:
                new_idxs.append(nodes.idx)
                if nodes.score > best_score:
                    best_nodes = nodes
        
        return new_idxs, best_nodes
        
        
    def _check_nodes(self, idx):
        "returns status, masked_nodes"
        #č co všechno se tu může dít?
        #č 1. nic se nezměnilo, všechny bodíky pořad patří k páru
        #č 2. Pozor, změna! Některé bodíky odpadly
        #č 3. Padla celá sada bodů.
        
        #č bodíky musí mít konečné souřadnice - ony prošly KD stromem
        #č vzorky - tím více
        nodes = self.nodes[idx]
        nodes_model = getattr(nodes, self.model_space)
        d3 = np.min(cdist(nodes_model, self.points[self._nsim:]), axis=1)
        mask = nodes.d2 <= d3
        
        if not np.any(mask):
            # -1 = 'outside', 0=success, 1=failure, 2=mix
            if nodes.event_id == 2:
                #č dáme zákazníkovi vědět před tím, než bodíky vyhodíme
                self.on_delete_mixed(idx)
            #č invalidace sady
            self.nodes.pop(idx)
            return 3, None
        elif np.all(mask):
            return 1, nodes
        else:
            del nodes.p_inside
            del nodes.p_fv
            del nodes.p_fw
            masked_nodes = nodes[mask]
            # update score
            self.score(masked_nodes)
            self.nodes[idx] = masked_nodes
            #č nejdřív update, teprve poté informujeme kontrahenta
            if nodes.event_id == 2:
                self.on_update_mixed(idx)
            return 2, masked_nodes
        
        ## checklist
        #self.couples[tp.couple_indices].append(nodes.idx)
        
        
    def _invalidate_couple(self, couple_indices):
        self.couples.pop(couple_indices)
        self.mixed_couples.pop(couple_indices, None)
        self.red_couples.pop(couple_indices, None)
        
    def explore_couple(self, couple_indices, nis):
        """
        Method is dedicated for the external usage.
        If someone thinks CV missed out some pair,
        it could say about and try to fix the thing.
        """
        i, j = max(couple_indices), min(couple_indices)
        couple_indices = (i, j)
        failsi = self.sample_box.failsi 
        if failsi[i]: #č první je červený
            if failsi[j]:
                #č červený kontakt
                # -1 = 'outside', 0=success, 1=failure, 2=mix
                event_id = 1
            else:
                #č žlutý kontakt
                # -1 = 'outside', 0=success, 1=failure, 2=mix
                event_id = 2
        
        else: #č první je zelený
            if failsi[j]:
                #č žlutý kontakt
                # -1 = 'outside', 0=success, 1=failure, 2=mix
                event_id = 2
            else:
                return 
                #оӵ Вож. Кулэ ӧвӧл
        
        tp = TwoPoints(self.points, couple_indices, workers=self.workers)
        #č pokud pár známe, jen přídáme další sadu bodů
        if couple_indices in self.couples:
            if event_id == 2:
                best_nodes = self.mixed_couples[couple_indices]
            else:
                best_nodes = self.red_couples[couple_indices]
            nodes = self.sample_alike(tp, best_nodes, nis)
            if nodes is not None:
                self.couples[couple_indices].append(nodes.idx)
                if nodes.score > best_nodes.score:
                    self._recommend_to_integration(nodes)
            
            return nodes
            
        #č pár nám není znám. Ještě jednou pokusíme nasypat body
        nodes = self.onboard_couple(tp, event_id=event_id, nis=nis)
        
        #č project páry na update
        for couple_indices in self._indices_to_update:
            #č nepodporuje slovník množinové operace
            self.update_couple(couple_indices)
        self._indices_to_update.clear()
        
        return nodes
        
        
        
        
        
    def onboard_couple(self, tp, event_id, nis):
        #č (i, j)? poprvé ta čísla vidíme
        i, j = couple_indices = tp.couple_indices
        
        midpoint = tp.half_point
        #č Nejdřív - zda je přímy kontakt, prostě na usečce
        #TData(X=nodes[gmask], imask, dd)
        nodes = self._store_masked(tp, midpoint, np.atleast_1d(np.inf), event_id)
        if nodes is not None:
            #č kontakt existuje, dáme vědět aktualizovačce 
            ##č (snad jediné místo kde tuhle funkci voláme)
            self._add_indices_to_update(i) #č může nás volat i explore_couple()
            self._add_indices_to_update(j) 
            self.couples[couple_indices].append(nodes.idx)
            
            
            ##nodes.w = 0
            ##nodes.p_couple = 0
            #č nsim score
            #č 0   -inf    #č myšleno hypoteticky
            #č 1   0       #č jednomu bodíku bych nedůveroval a nebral bych největší p
            #č 2   ~p_couple
            #č chcu, aby score byl měnší jak u připadného skutečného bodíku.
            nodes.score = -1 #č není špatný score
            
            #č chcu zjednodušit logiku. 
            #č Pokud máme pouze jeden bodík, tak snad i bude nejlepším?
            #č Nechť veškeré couples v třídě budou konzistentní.
            #č dočasně nejlepší tenhlensten, tak ho doporučíme k integraci
            self._recommend_to_integration(nodes)
        
        
        #č a teď deme vzorkovat
        #vec = couple_nodes[0] - couple_nodes[1]
        vec = tp.normal_vector
        r = np.sqrt(np.inner(vec, vec)) / 2
        
        # initially we'll perform IS from multivariate Gaussian distribution 
        # with center at mid node (i. e. in beetween of two points)
        # and single standard deviation (let's say "ball" sampling)
        #č r-ko dělíme base_r'kem k získání sigmy.
        sampling_plan_N, h_pdf = IS_stat.get_norm_plan(nis, self.ndim, \
                                mean=midpoint, 
                                std=r/self.base_r, design=None)
        
        nodes = self._store_masked(tp, sampling_plan_N, h_pdf, event_id)
        if nodes is None:
            return None
        
        #č kontakt existuje, dáme vědět aktualizovačce 
        ##č (druhé místo kde tuhle funkci voláme)
        self._add_indices_to_update(i) #č může nás volat i explore_couple()
        self._add_indices_to_update(j) 
        self.couples[couple_indices].append(nodes.idx)
            
        #č jdeme do adaptivního IS vzorkování
        #č zde jenom vzorkujeme; 
        #č integrování, vyhodnocování je jinde a později!
        best_score = -np.inf
        while nodes.score > best_score:
            best_nodes = nodes
            best_score = nodes.score
            nodes = self.sample_alike(tp, nodes, nis)
            self.couples[couple_indices].append(nodes.idx)
            
        
        #č doporučíme k integraci
        #č Má cenu integrovat pouze nejreprezentativnější sadu bodů.
        #č Špatné sady je velmi obtižné správně započítavat
        #č aby nezhoršovaly výsledek.
        #č Možná TrueIS by si mohl s tím poradit, ale 
        #č zde je to totálná zbytečnost
        # -1 = 'outside', 0=success, 1=failure, 2=mix
        self._recommend_to_integration(best_nodes)
        return best_nodes
            
        
        
    def score(self, nodes):
        #č p_couple není finální pravděpodobnost
        #č bodíky ještě musejí projít konvexní obálkou
        nodes.p_couple = np.sum(nodes.w)
        #č nsim score
        #č 0   -inf    #č myšleno hypoteticky
        #č 1   0       #č jednomu bodíku bych nedůveroval a nebral bych největší p
        #č 2   ~p_couple
        nodes.score = nodes.p_couple * np.log(len(nodes))
        
        
    def sample_alike(self, tp, nodes, nis):
        """
        č zde jenom vzorkujeme; 
        č integrování, vyhodnocování je jinde a později!
        returns: node_idx, score
        """
        assert len(nodes) > 0
        
        nodes_model = getattr(nodes, self.model_space)
        
        
        if len(nodes) > 1:
            #č potřebujeme vahy, pokud bodíků víc jak jeden. Takže v cajku.
            sampling_plan, h_pdf = IS_stat.sample_alike(nodes_model, \
                                            weights=nodes.w, nis=nis, d=self.d)
            
        else:
            #č máme jeden bodík. To znaméná, že kontakt existuje
            #č a zákazník hodla provést vzorkování.
            #č Pro tento vzácný případ naše společnost nabízí řešení - 
            #č zapojít informace o těch dvou vzorkách, ke kterým patří.
            #č Na ty dva vzorky se nemůžeme spoléhat z hlediska směrových 
            #č preferencí, použijme je jen jako zdroj jakési 
            #č charakteristické délky. Těžiště necháme v tamtamtom bodíku.
            
            #č d2 je vždy přítomný. Zajistěno v _store_masked()
            #č Specifikem CandyNodes je, že d2 pro jeden bodík může vrátit
            #č jak skalarní hodnotu, tak i pole s jediným prvkem
            #č numpy si poradí s obojí
            r = nodes.d2
            
            # we'll perform IS from multivariate Gaussian distribution 
            # centered at the isolated node 
            # and some standard deviation
            #č r-ko dělíme base_r'kem k získání sigmy.
            sampling_plan, h_pdf = IS_stat.get_norm_plan(nis, self.ndim,
                                    mean=nodes_model[0], 
                                    std=r/self.base_r, design=None)
                                
                                
        return self._store_masked(tp, sampling_plan, h_pdf, nodes.event_id)
        
        
        
    def _store_masked(self, tp, nodes_model, h_pdf, event_id):
        """ returns nodes, mask, tdata
        
        č logika celé třídy je, že máme-li příslušné kontaktu bodíky,
        č tak stojí za to je uložit. 
        č Proto metod maskuje a je-li co ukladat tak i rovnou uloži 
        č a vratí jejich index (jinak vratí nulu).
        č To ale vyžaduje trochu větší pozornost - 
        č bodíky pak někdo má odebrat a invalidovat.
        č Proto i podtržitko v názvu
        """
        #TData(X=nodes[gmask], imask, dd)
        status, mask, tdata = tp.query(nodes_model)
        
        if not status:
            return None
            
            
        #č přídat, zavolat kólbek
        f = self.sample_box.f_model
        #č tady je to super, ťažké datové struktury f-modelu
        #č vytvaříme na jíž vyfiltrovaných datéch.
        nodes = f.new_sample(tdata.X, self.model_space)
        couple_stats = {'event_id': event_id, 'couple': tp.couple_indices}
        nodes = CandyNodes(nodes, couple_stats, d2=tdata.dd)
        idx = self.nodes.add(nodes)
        nodes.idx = idx
        
        w = nodes.pdf(self.model_space) / h_pdf[mask] / len(mask)
        #č Filtrace stromem (která pořad je) zaručuje, 
        #č že bodíky (přenejmenším na vstupu f_modelu) 
        #č mají konečné souřadnice bez nan a nekonečno.
        #č To ale nic neříká o hustotách 
        nodes.w = w * np.isfinite(w)
        #č beztak skorem v této třídě projde každá sada vzorků 
        self.score(nodes)
        
        #č červené kontakty prostě rovnou integrujeme
        #č takže netřeba pro ně ukladat zbytečná data
        # -1 = 'outside', 0=success, 1=failure, 2=mix
        if event_id == 2:
            nodes.imask = tdata.imask
            self.on_add_mixed(nodes)
        
        
        ## checklist
        #č _store_masked dělá na urovní nodes
        #č zbytek nechť dělá kód vejš
        #č michá se to u kontrol párů z minule
        #self.couples[tp.couple_indices].append(nodes.idx) 
        #self.mixed_couples #č v této fázi nevíme,
        #self.red_couples   #č zda je sada ta nejlepší
        #self._add_indices_to_update(j) #č vždyť já nevím, jestli se jedná o aktualizaci
        
        return nodes
        
        
        
    #č tahle funkce může být volána i skřiňkou
    def estimate_mixed(self, nodes):
        "assigns d1, fv, fw"
        nodes_model = getattr(nodes, self.model_space)
        i, j = nodes.couple
        imask = nodes.imask
        d2 = nodes.d2
        
        di1 = cdist(nodes_model[imask], np.atleast_2d(self.points[i])).reshape(-1)
        dj1 = cdist(nodes_model[~imask], np.atleast_2d(self.points[j])).reshape(-1)
        
        d1 = np.empty(len(imask))
        d1[imask] = di1
        d1[~imask] = dj1
        #č skříňka bude chtit d1
        nodes.d1 = d1
        
        Pdf_i = self.PDF[i]
        Pdf_j = self.PDF[j]
        
        if self.failsi[i]:
            df[imask] = di1
            df[~imask] = d2[~imask]
            dpdf = df * Pdf_i
        else:
            df[imask] = d2[imask]
            df[~imask] = dj1
            dpdf = df * Pdf_j
        
        sumpdf = np.empty(len(imask))
        sumpdf[imask] = di1 * Pdf_i + d2[imask] * Pdf_j
        sumpdf[~imask] = dj1 * Pdf_j + d2[~imask] * Pdf_i
        
        w = nodes.w
        
        nodes.fv = w * df / (d1 + d2)
        nodes.fw = w * dpdf / sumpdf
        
        
        
        
        
    def get_pf_estimation(self):
        #č zkusme představit, 
        #č že někdo bude volat tuhle funkci pořad dokola,
        #č aniž by se přídaly nové vzorky.
        self._update()
        failure, mixed, pfv, pfw = 0
        
        valid_insides = self._valid_outsides
        is_inside = self.hull.is_inside
        
        for nodes in self.red_couples.values():
            if nodes.idx not in valid_insides:
                nodes.inside = mask = is_inside(nodes)
                valid_insides.add(nodes.idx)
                nodes.p_inside = p_inside = np.sum(nodes.w[mask])
            else:    #č kód třídy musí zajistit konzistentnost p_inside, 
                try: #č mazajic ji kdyby něco 
                    p_inside = nodes.p_inside
                except AttributeError:
                    nodes.p_inside = p_inside = np.sum(nodes.w[nodes.inside])
                    
            failure += p_inside
            pfv += p_inside
            pfw += p_inside
        
        
        for nodes in self.mixed_couples.values():
            if not hasattr(nodes, 'fw'):
                self.estimate_mixed(nodes)
                
            if nodes.idx not in valid_insides:
                nodes.inside = mask = is_inside(nodes)
                valid_insides.add(nodes.idx)
                nodes.p_inside = p_inside = np.sum(nodes.w[mask])
                nodes.p_fv = p_fv = np.sum(nodes.fv[mask])
                nodes.p_fw = p_fw = np.sum(nodes.fw[mask])
            else:    #č kód třídy musí zajistit konzistentnost peček, 
                try: #č mazajic je kdyby něco 
                    p_inside = nodes.p_inside
                    p_fv = nodes.p_fv
                    p_fw = nodes.p_fw
                except AttributeError:
                    mask = nodes.inside
                    nodes.p_inside = p_inside = np.sum(nodes.w[mask])
                    nodes.p_fv = p_fv = np.sum(nodes.fv[mask])
                    nodes.p_fw = p_fw = np.sum(nodes.fw[mask])
            
            mixed += p_inside
            pfv += p_fv
            pfw += p_fw
        
        return failure, mixed, pfv, pfw
        
        



#!/usr/bin/env python
# coding: utf-8

import numpy as np
import numpy.ma as ma
import scipy.stats as stats

from scipy.spatial import cKDTree
from scipy.spatial import Delaunay
from scipy import spatial
from scipy import interpolate

import collections # for defaultdict

from . import f_models
from . import simplex as six
from . import IS_stat
from .IS_stat import IS
from .candybox import CandyBox
from . import sball 



#
# l2 Voronoi cKDTree IS localization
#
 
# Rbf_estimation so far
def L2_Voronoi_cKDTree(sample_box, nis=50000):    
    nsim, nvar = np.shape(sampled_plan_R)
    # Estimation of pf given a list of red and green points (Voronoi)
    
    sampled_plan_R_ma = np.ma.asarray(sampled_plan_R)
    
    # tady bych nedaval Rd 
    #grad = np.abs(gradient([0 for j in range(nvar)]))
    tree = cKDTree(sampled_plan_R)
    

    rbf_func = interpolate.Rbf(sampled_plan_R[:,0],sampled_plan_R[:,1],Z)

    Rbf_estimation = 0 # inicializace
    L2_Voronoi_cKDTree_test = 0
    #failure_points_indexes = np.argwhere(failsi)
    
    points_probabilities = np.empty(nsim) # probabilities of partial (failure) event
    
    # loop over points (need to integrate red regions)
    for i in range(nsim): # loop over failing points only
        
        sampled_plan_R_ma.mask = ma.nomask
        sampled_plan_R_ma[i] = ma.masked
        
        # neosvědčílo se
#==============================================================================
#         delta_Rd_matrix = sampled_plan_Rd_ma - sampled_plan_Rd[i]
#         mindist = [np.min(np.abs(np.where(delta_Rd_matrix[:,j] < 0,delta_Rd_matrix[:,j], f_i[j].std() ))) + np.min(np.abs(np.where(delta_Rd_matrix[:,j] > 0,delta_Rd_matrix[:,j], f_i[j].std() ))) for j in range(nvar)]
#                                          
#         # set the minimum distance as the standard deviation of IS densisty
#         h_i = [stats.norm(sampled_plan_Rd[i,j], mindist[j] ) for j in range(nvar)] #! dosadit  standard deviation pddle chutí
#==============================================================================
        # find distance to the nearest sampling point (from all points)
        mindist = np.min(np.sum(np.square(sampled_plan_R_ma - sampled_plan_R[i]), axis=1))**0.5
                                 
        # set the minimum distance as the standard deviation of IS densisty
        h_i = [stats.norm(sampled_plan_R[i,j], 2*mindist ) for j in range(nvar)] #! dosadit  standard deviation pddle chutí
                                 
        # use IS sampling density with center equal to the current "red" point
                                 
        # select nis = 100 points from IS density and 
        # if the point has its nearest neighbor any red point from the sampled_plan, 
                                 
        h_plan = np.zeros((nis, nvar))
        for j in range(nvar):
            h_plan[:, j] = h_i[j].ppf(np.random.random(nis)) # realizace váhové funkce náhodné veličiny
    
        # Rozptyl corrected IS
        weights_sim = np.prod([f_i[j].pdf(h_plan[:, j]) / h_i[j].pdf(h_plan[:, j]) for j in range(nvar)], axis=0) # [f1/h1, ..., fn/hn]
    
       
    
        dd, ii = tree.query(h_plan)
    
        Vor_mask = np.where(ii==i, True, False)
        
        zm = rbf_func(h_plan[:, 0][Vor_mask], h_plan[:, 1][Vor_mask])        
        
        points_probabilities[i] = np.sum(weights_sim[Vor_mask][zm<0]) / nis
        L2_Voronoi_cKDTree_test += np.sum(weights_sim[Vor_mask]) / nis
    
    
    Rbf_estimation = np.sum(points_probabilities)
    
    return Rbf_estimation, L2_Voronoi_cKDTree_test
    





# cKDTree
"""
# dd  - The distances to the nearest neighbors 
# ii - The locations of the neighbors in self.data
# k - The list of k-th nearest neighbors to return. 
# If k is an integer it is treated as a list of [1, … k] (range(1, k+1)). 
# Note that the counting starts from 1
# p - Which Minkowski p-norm to use. 
# 1 is the sum-of-absolute-values “Manhattan” distance 2 is the usual Euclidean distance 
# infinity is the maximum-coordinate-difference distance 

dd, ii = tree.query(h_plan_model, k=1, p=p_norm)
"""


#
# Voronoi estimation
#    (gradient tu nechcu)
def Voronoi_tesselation(sample_box, model_space='Rn', sampling_space=None, p_norm=1, budget=20000, callback=None):
    """
    Voronoi estimations 
    budget=20000
    """
    global_stats = {'Voronoi_failure_rate':0}

    
    
    nsim = sample_box.nsim
    
        
    # jsou to informace nejen pro callback 
    # nis upresníme dále
    estimation={'method': "Voronoi_tesselation",  'p_norm':p_norm, 'nis':0, 'nsim':nsim}
    estimation['model_space'] = model_space
    estimation['sampling_space'] = sampling_space
    
    if len(sample_box.failure_points) > 0:
        nis = max(round(budget/len(sample_box.failure_points)), 100)
        estimation['nis'] = nis
    
        # vytahneme ze sample_boxu rozdělení
        f = sample_box.sampled_plan
        
        # já vím, že sample box pokážde failure_points přepočítavá
        failure_points = sample_box.failure_points 
        
        # zde provadím rozdělení na prostor, ve kterém vzorkujem
        # a prostor "modelu", vô ktôrom, v podstatě, měříme vzdaleností
        sampled_plan_model = getattr(sample_box, model_space)
        tree = cKDTree(sampled_plan_model)
        
        
        if sampling_space is None:
            sampling_space = model_space
            # sing like sampling
            sampled_plan_sing = sampled_plan_model
            tree_sampling = tree
        else:
            sampled_plan_sing = getattr(sample_box, sampling_space)
            # narychlo, moc to nepomůže, neboť asi po 16 vzorcích počítá brut forsem 
            tree_sampling = cKDTree(sampled_plan_sing,  compact_nodes=False, balanced_tree=False)
            
        # find distance to the nearest sampling point (from all points)
        dd2, ii2 = tree_sampling.query(sampled_plan_sing[failure_points], k=[2], p=p_norm)
        mindist_sing = dd2.reshape(-1)
        #ii2 = ii2.reshape(-1)
        
        
        
        # loop over points (need to integrate red regions)
        for i, i2 in zip(sample_box.failure_points, range(len(mindist_sing))): # loop over failing points only
            # use IS sampling density with center equal to the current "red" point
            # set the minimum distance as the standard deviation of IS densisty
            h_i = [stats.norm(sampled_plan_sing[i,j], 1.5*mindist_sing[i2]) for j in range(sample_box.nvar)] #! dosadit  standard deviation podle chutí
            h = f_models.UnCorD(h_i)
            
                                     
            # select nis = 100 points from IS density 
            # sice to má nazev h_plan, ale nese rozdělení a hustoty v f-ku
            h_plan = IS(f, h, space_from_h='R', space_to_f=sampling_space, Nsim=nis) 
            
            # součet váh nemá cenu kontrolovat, jedná tam nebude, nevyjde
            
            
            h_plan_model = getattr(h_plan, model_space)
            dd, ii = tree.query(h_plan_model, k=1, p=p_norm)
            
            cell_probability = np.sum(h_plan.w[ii==i]) / nis
            global_stats['Voronoi_failure_rate'] += cell_probability
            
            # kolbek ↓
            if callback is not None:
                cell_stats = {'Voronoi_failure_rate': cell_probability}
                cell_stats['cell_probability'] = cell_probability
                cell_stats['event'] = 'failure'
                callback(estimation=estimation, nodes=h_plan[ii==i], cell_stats=cell_stats, out_nodes=h_plan[ii!=i])

        
    result = {**estimation, **global_stats}
    # try to put it to the BlackBox
    try:
        sample_box.estimations.append(result)
    except: # asi ani nebudu neuspěch hlasit
        pass
        
    return result
    

    
    
    
#
# Voronoi 2_point estimation
# 
def Voronoi_2_point_estimation(sample_box, model_space='Rn', sampling_space=None, p_norm=1, gradient=None, budget=20000, callback=None):
    """
    Voronoi_2_point estimations 
    budget=20000
    """
    
    # tak, na rovinu
    # нет ножек - нет мультиков
#    if gradient is None:
#        return Voronoi_2_point_cKDTree(sample_box, model_space=model_space,sampling_space=sampling_space,\
#                                                            p_norm=p_norm, budget=budget, callback=callback)

#    if gradient is None:
#        gradient = lambda *_, **__: np.ones(sample_box.nvar)

    if callback is None:
        callback = lambda *_, **__: None
    
    nsim = sample_box.nsim
    # ощень, ощень скромненько
    p_base = 0.3 # chcu, aby behem první iterace p_base tečiček jistě dopadlo do buňky
    nis_base = int((sample_box.nvar+1)/p_base)
    nis = int(max(budget/nsim, nis_base / stats.binom.sf(int(sample_box.nvar+1), nis_base, p_base)))
        
    # jsou to informace pro callback 
    estimation={'method': "Voronoi_2_point_estimation",  'p_norm':p_norm, 'nis':nis, 'nsim':nsim}
    estimation['model_space'] = model_space
    estimation['sampling_space'] = sampling_space
    
    
    # vytahneme ze sample_boxu rozdělení
    f = sample_box.sampled_plan
    
    # já vím, že sample box pokážde failsi přepočítavá
    failsi = sample_box.failsi 
    
    PDF = sample_box.pdf(model_space)
    
    # zde provadím rozdělení na prostor, ve kterém vzorkujem
    # a prostor "modelu", vô ktôrom, v podstatě, měříme vzdaleností
    sampled_plan_model = getattr(sample_box, model_space)
    tree = cKDTree(sampled_plan_model)
    
    
    if sampling_space is None:
        sampling_space = model_space
        # sing like sampling
        sampled_plan_sing = sampled_plan_model
        tree_sampling = tree
    else:
        sampled_plan_sing = getattr(sample_box, sampling_space)
        # narychlo, moc to nepomůže, neboť asi po 16 vzorcích počítá brut forsem 
        tree_sampling = cKDTree(sampled_plan_sing,  compact_nodes=False, balanced_tree=False)
        
    # find distance to the nearest sampling point (from all points)
    dd2, ii2 = tree_sampling.query(sampled_plan_sing, k=[2], p=p_norm)
    mindist_sing = dd2.reshape(-1)
    
    # chcu, aby behem první iterace pulka (p_base) tečiček jistě dopadla do buňky
    base_r = sball.Sball(sample_box.nvar).get_r(1 - p_base)
    # np.sum(abs(stats.norm(0, 10).rvs(1000))<sb.get_r(.5)*10)
    
    

    global_stats = collections.defaultdict(int)
    
    
    
    # zde postupně v cyklu prochazíme všemi body (tj. vzorky)
    # a omezujeme se pouse nejbližšími bodíkama
    # tynhlenstím zajišťujeme disjunktnost 
    # a môžeme všechny nasbírané pravděpodobnosti jednoduše sčítat
    for i in range(nsim): # loop over all points, not failing points only
    
        # use IS sampling density with center equal to the current point
        # set the minimum distance as the standard deviation of IS densisty
        # mindist dělíme dvěma - hranice buňky je na půlcesty
        # a dělíme base_r'kem k získání sigmy. Přemejmenším empiricky mně to vychází
        # (u Norm'u vždy zadaváme směrodatnou odchylku)
        h_i = [stats.norm(sampled_plan_sing[i,j], mindist_sing[i]/2/base_r) for j in range(sample_box.nvar)] 
        h = f_models.UnCorD(h_i)
        
                                 
        # select nis points from IS density 
        # sice to má nazev h_plan, ale nese rozdělení a hustoty v f-ku
        h_plan = IS(f, h, space_from_h='R', space_to_f=sampling_space, Nsim=nis) 
        
        # součet váh nemá cenu kontrolovat, jedná tam nebude, nevyjde
        
        h_plan_model = getattr(h_plan, model_space)
        dd, ii = tree.query(h_plan_model, k=1, p=p_norm)
        
        # nechám s velkým písmenkem
        Vor_mask = ii==i
        h_plan_model_ma = h_plan_model[Vor_mask]
        weights_sim = h_plan.w[Vor_mask]
        # dd1 jsou vzdalenosti tečiček do centra Voroneho buňky
        dd1 = dd[Vor_mask]
        
        trace = 0
        nis_ma = len(weights_sim)
        
        h_plan_sing_ma = getattr(h_plan, sampling_space)[Vor_mask]
        S_bc = np.cov(h_plan_sing_ma, rowvar=False, bias=True, aweights=weights_sim)
        barycenter = np.average(h_plan_sing_ma, axis=0, weights=weights_sim)
        # effective nis
        nis_eff = nis
        #print("start", i)
        
        while trace < np.trace(S_bc):
            # number of nodes inside the cell matters too
            trace = np.trace(S_bc)# * nis_ma
            #print(S_bc)
            #print(nis_ma)
        
            
            
            
            # matika
            w, v = np.linalg.eig(S_bc)
            
            # use IS sampling density with center equal to the cell's barycenter
            # set the minimum distance as the standard deviation of the IS densisty
            # u stats.norm zadáváme směrodatnou odchylku, to je asi správné
            sigmas = np.sqrt(w) / base_r #(sample_box.nvar+2) #! dosadit  standard deviation podle chutí
            h_i = [stats.norm(0, sigmas[j]) for j in range(sample_box.nvar)]
            # rozdělení ve vlastním prostoru
            # select nis = 100 points from IS density 
            h_L = f_models.UnCorD(h_i)(nis)
            
            # здесь уже так легко не отделаемся. Трансформовать кароно.
            h_plan_bc = (v @ h_L.R.T).T
            h_plan_sing = h_plan_bc + barycenter
            
            # sice to má nazev h_plan, ale nese rozdělení a hustoty v f-ku
            h_plan_part = f.new_sample(h_plan_sing, sampling_space)
            
            
            # jdeme na ty hustoty
            # mně příjde, že je to legalní
            # sice samply podporujou maskovaní, to je ale drahé
            weights_sim_part = h_plan_part.pdf(sampling_space) / h_L.pdf('R') # snad je to správně
            h_plan.add_sample(CandyBox(h_plan_part, w=weights_sim_part))
            
            # vyfiltrujeme vzorky
            h_plan_model_part = getattr(h_plan_part, model_space)
            dd, ii = tree.query(h_plan_model_part, k=1, p=p_norm)
            
            # parta
            Vor_mask_part = ii==i
            weights_sim_part = weights_sim_part[Vor_mask_part]
            
            nis_ma = len(weights_sim_part)
            
            # zajišťovat Vor_mask je docela zbytečně, je to jen pro out_nodes,
            # které se zatím nikdě nepouživá
            Vor_mask = np.append(Vor_mask, Vor_mask_part)
            
            
            
            
            h_plan_model_ma = np.vstack((h_plan_model_ma, h_plan_model_part[Vor_mask_part]))
            weights_sim = np.append(weights_sim, weights_sim_part)
            # dd1 jsou vzdalenosti tečiček do centra Voroneho buňky
            dd1 = np.append(dd1, dd[Vor_mask_part])
            
            # zkusím těžiště počitat jen pro partu - možná tak algoritmus bude agresivnější?
            #barycenter = np.average(h_plan_sing[Vor_mask_part], axis=0, weights=weights_sim_part)
            h_plan_sing_ma = np.vstack((h_plan_sing_ma, h_plan_sing[Vor_mask_part])) 
            #S_bc = np.cov(h_plan_sing[Vor_mask_part], rowvar=False, bias=True, aweights=weights_sim_part)
            S_bc = np.cov(h_plan_sing_ma, rowvar=False, bias=True, aweights=weights_sim)
            barycenter = np.average(h_plan_sing_ma, axis=0, weights=weights_sim)
            nis_eff += nis
            
        
        
        #print(S_bc)
        #print(nis_ma)
        
        
        
        
        cell_stats = dict()
        # musí sa (na konci) rovnat jedne
        # opravdu dělíme nis'em, jako v normálním IS
        # nikoliv počtem příjatých bodíků, 
        # protože průměrná vaha je o hodně mene významná metrika
        cell_stats['cell_probability'] = np.sum(weights_sim) / nis_eff
        
        
        # tu bacha!
        # takhle se počíta, pokud se netrapíme gradijentem
        # a je to trošiňku optimizovaný, takže čert se nevyzná
        if gradient is None:
            # indexy ii nás moc nezajimajou
            # vzdalenosti snad byjsme zvladli použit?
            
            
            dd2, ii2 = tree.query(h_plan_model_ma, k=[2], p=p_norm)
            dd2 = dd2.reshape(-1)
            ii2 = ii2.reshape(-1)
            
            # tahle hračka s indexy je pro numpy poměrně drahá
            failsii_2 = failsi[ii2]
            # jeden vzorek (včetně hustoty PDF[i])  je nám vždy znám
            # porucha
            if failsi[i]:
                points_1 = PDF[i] * dd2
                node_pf_estimations = points_1 / (points_1 + PDF[ii2] * dd1)
                node_pf_estimations = np.where(failsii_2,1, node_pf_estimations)
                node_pf_pure_estimations = dd2 / (dd1 + dd2)
                node_pf_pure_estimations = np.where(failsii_2,1, node_pf_pure_estimations)
            
                cell_stats['Voronoi_2_point_upper_bound'] = cell_stats['cell_probability'] 
                cell_stats['Voronoi_2_point_failure_rate'] = np.sum(weights_sim * node_pf_estimations) / nis_eff
                cell_stats['Voronoi_2_point_pure_failure_rate'] = np.sum(weights_sim * node_pf_pure_estimations) / nis_eff
                cell_stats['Voronoi_2_point_lower_bound'] = np.sum(weights_sim[failsii_2]) / nis_eff
                cell_stats['Voronoi_failure_rate'] = cell_stats['cell_probability'] 
                nodes=CandyBox(h_plan.sampling_plan[Vor_mask], w=weights_sim, node_pf_estimations=node_pf_estimations,\
                                    node_pf_pure_estimations=node_pf_pure_estimations, dd1=dd1, dd2=dd2, ii2=ii2)
            
            # neporucha    
            else:
                dd1 = dd1[failsii_2]
                dd2 = dd2[failsii_2]
                points_1 = PDF[i] * dd2
                points_2 = PDF[ii2[failsii_2]] * dd1
                
                node_pf_estimations = points_2 / (points_1 + points_2)
                node_pf_pure_estimations = dd1 / (dd1 + dd2)
            
                cell_stats['Voronoi_2_point_upper_bound'] = np.sum(weights_sim[failsii_2]) / nis_eff
                cell_stats['Voronoi_2_point_failure_rate'] = np.sum(weights_sim[failsii_2]*node_pf_estimations) / nis_eff
                cell_stats['Voronoi_2_point_pure_failure_rate'] = np.sum(weights_sim[failsii_2] * node_pf_pure_estimations) / nis_eff
                cell_stats['Voronoi_2_point_lower_bound'] = 0
                cell_stats['Voronoi_failure_rate'] = 0
                nodes=CandyBox(h_plan.sampling_plan[Vor_mask][failsii_2], w=weights_sim[failsii_2], node_pf_estimations=node_pf_estimations,\
                                            node_pf_pure_estimations=node_pf_pure_estimations, dd1=dd1, dd2=dd2, ii2=ii2[failsii_2])
                
                # take something with corresponding length
                zeros = np.zeros(len(weights_sim) - len(dd2))
                # add remaining nodes
                nodes.add_sample(CandyBox(h_plan.sampling_plan[Vor_mask][~failsii_2], w=weights_sim[~failsii_2], node_pf_estimations=zeros,\
                                            node_pf_pure_estimations=zeros, ii2=ii2[~failsii_2]))
        
        
        # takhle - pokud chceme gradient použit
        # je třeba eště zoptimalizovať
        else:
            # kolik bodíků jsou nejbližší k mému vzorečkovi
            # np.empty() implicitně má dtype=float
            # tyhle blbosti ponechám jen kvůli callbackovi
            node_pf_estimations = np.empty(len(h_plan_model_ma))
            node_pf_pure_estimations = np.empty(len(h_plan_model_ma))# pure distance estimation
            node_failsi = np.empty(len(h_plan_model_ma), dtype=bool) # for L1 Voronoi # co to je za L1 Voronoi?
            
            # projdeme přes každej bodíček
            for node_idx in range(len(h_plan_model_ma)):
                # KDTree byl použit jen k rozdělení na disjunktní úseky, veškerej děj se odehravá tu
                # a to je všechno kvůli gradientu
                node = h_plan_model_ma[node_idx]
                # axis=1 - sčítá všechy směry dohromady, vysledkem je 1D pole rozměru nsim 
                inode2points_model_matrix = np.sum(np.abs(((sampled_plan_model - node) * gradient(node))**p_norm), axis=1)
                #print(inode2points_Rd_matrix)
                
                """
                partition - 
                Creates a copy of the array with its elements rearranged in such a way that
                 the value of the element in k-th position is in the position it would be in a sorted array. 
                 All elements smaller than the k-th element are moved before this element 
                 and all equal or greater are moved behind it. The ordering of the elements in the two partitions is undefined.
                """
                idx = np.argpartition(inode2points_model_matrix, 1) # musí tu bejt 1, počítá sa od nuly
                # je to správný, neboť numpy zaručuje, že druhej prvek (s indexem 1) bude na druhem místě
                node_failsi[node_idx] = failsi[idx[0]] 
                
                                          
                points_weight = PDF[idx[0]] / inode2points_model_matrix[idx[0]] + PDF[idx[1]] / inode2points_model_matrix[idx[1]]
                points_distances = 1 / inode2points_model_matrix[idx[0]] + 1 / inode2points_model_matrix[idx[1]]
                
                failure_weight  = int(failsi[idx[0]]) * PDF[idx[0]] / inode2points_model_matrix[idx[0]]
                failure_weight += int(failsi[idx[1]]) * PDF[idx[1]] / inode2points_model_matrix[idx[1]]
                
                failure_distance = int(failsi[idx[0]]) / inode2points_model_matrix[idx[0]] + int(failsi[idx[1]]) / inode2points_model_matrix[idx[1]]
                
                
                node_pf_estimations[node_idx] = failure_weight/points_weight
                node_pf_pure_estimations[node_idx] = failure_distance/points_distances
            
            
            cell_stats['Voronoi_2_point_upper_bound'] = np.sum(h_plan.w[Vor_mask]*np.ceil(node_pf_estimations)) / nis_eff
            cell_stats['Voronoi_2_point_failure_rate'] = np.sum(h_plan.w[Vor_mask]*node_pf_estimations) / nis_eff
            cell_stats['Voronoi_2_point_pure_failure_rate'] = np.sum(h_plan.w[Vor_mask]*node_pf_pure_estimations) / nis_eff
            cell_stats['Voronoi_2_point_lower_bound'] = np.sum(h_plan.w[Vor_mask]*np.floor(node_pf_estimations)) / nis_eff
            cell_stats['Voronoi_failure_rate'] = np.sum(h_plan.w[Vor_mask]*node_failsi) / nis_eff
            
            nodes=CandyBox(h_plan.sampling_plan[Vor_mask], w=h_plan.w[Vor_mask], node_pf_estimations=node_pf_estimations,\
                                            node_pf_pure_estimations=node_pf_pure_estimations, node_failsi=node_failsi)
            
            
        
        for key, value in cell_stats.items():
            global_stats[key] += value
        
        # kolbek ↓
        if failsi[i]:
            cell_stats['event'] = 'failure'
        else:
            cell_stats['event'] = 'success'
        callback(estimation=estimation, nodes=nodes, cell_stats=cell_stats, out_nodes=h_plan[~Vor_mask])

        
    
    result = {**estimation, **global_stats}
    # try to put it to the BlackBox
    try:
        sample_box.estimations.append(result)
    except: # asi ani nebudu neuspěch hlasit
        pass
        
    return result
    
    # dedictvi
#for k in range(len(ii)):
#    points_weigths[ii[k]] = points_weigths[ii[k]] + weights_sim[k] / nis
#    near_neighbors[ii[k]] = near_neighbors[ii[k]] + 1
#    Vor_mask[k] = failsi[ii[k]]    
    
    

    
    



def _issi_estimate_outside(f, model_space, sampling_space='G', \
                 powerset_correction=True, design=None, callback=None, outside_budget=1000):
    #č zde f-ko musí taky obsahovat vzorky!
    
    shull = six.Shull(f, model_space, sampling_space, \
                      powerset_correction, incremental=False, design=design)
    
    nodes = shull.integrate(outside_budget)
    
    # -1 = 'out'
    # -2 = 'inside'
    #č převezmeme ISSI
    oiss = shull.oiss
    
    
    # events přepíšeme
    weighted_mean, __, events = oiss.get_means()
    # outside probability
    #č nevyrovnané!
    outside_probability = weighted_mean[events==-1][0]
    
    if callback is not None:
        cell_stats = dict()
        #č tohle je opravdu jen pro formu
        cell_stats['cell_probability'] = outside_probability
        cell_stats['vertex_estimation'] = 0
        cell_stats['weighted_vertex_estimation'] = 0
        cell_stats['event'] = 'outside' #č "outside" se mi libí víc než prostě "out"
        
        
        #:) kolbek ↓
        mask = nodes.is_outside
        #nodes = nodes[mask]
        # out_nodes here meant "not needed"
        out_nodes = nodes[~mask] #č možná je čas se zbavit toho, co se stejně nepouživá?
        callback(shull=shull, nodes=nodes[mask], cell_stats=cell_stats, out_nodes=out_nodes) 
        
    return oiss, outside_probability

    


def full_simplex_estimation(sample_box, model_space='Rn', sampling_space=None, weighting_space=None,\
                           outside_budget=1000, simplex_budget=100, design=None, callback=None):
    """
    Delaunay triangulation
    """
    
    #č zde provadím rozdělení na prostor, ve kterém vzorkujem
    #čs a prostor "modelu", vô ktôrom, v podstatě, měříme vzdaleností
    if sampling_space is None:
        sampling_space = model_space
    
    if weighting_space is None:
        weighting_space = model_space
        
    
    #č jsou to informace (nejen) pro callback 
    estimation={'method': "full_simplex_estimation", 'nsim':sample_box.nsim}
    estimation['model_space'] = model_space
    estimation['sampling_space'] = sampling_space
    estimation['weighting_space'] = weighting_space
    estimation['outside_budget'] = outside_budget
    estimation['simplex_budget'] = simplex_budget
    estimation['design'] = str(design)
    
    
    #č vytahneme ze sample_boxu rozdělení
    f = sample_box.f_model
    
    # powerset_correction=False
    #č totíž my zodpovidáme za to, že Triangulaci pošleme správné ISSI 
    issi, outside = _issi_estimate_outside(f, model_space,\
                     sampling_space=sampling_space, powerset_correction=False,\
                design=design, callback=callback, outside_budget=outside_budget)
                
    tri = six.FullSamplingTriangulation(sample_box, tri_space=model_space, issi=issi, \
             weighting_space=weighting_space, incremental=False,\
              on_add_simplex=callback, sampling_space=sampling_space, \
              simplex_budget=simplex_budget, design=design)
              
    tri.integrate()
                
    # TRI-compatible estimation
    # -1 = 'outside', 0=success, 1=failure, 2=mix
    estimations = tri.get_pf_estimation()
    #č doplním
    estimations['global_stats']['outside'] = outside
    
    result = {**estimation, **estimations}
    # try to put it to the BlackBox
    try:
        sample_box.estimations.append(result)
    except: #č asi ani nebudu neuspěch hlasit
        pass
        
    return result
        




def fast_simplex_estimation(sample_box, model_space='Rn', sampling_space=None, weighting_space=None,\
                                            outside_budget=1000, simplex_budget=100, design=None, callback=None):
    
    
    #č vytahneme ze sample_boxu rozdělení
    f = sample_box.f_model
    
    
    #č zde provadím rozdělení na prostor, ve kterém vzorkujem
    #čs a prostor "modelu", vô ktôrom, v podstatě, měříme vzdaleností
    sampled_plan_model = getattr(sample_box, model_space)
    if sampling_space is None:
        sampling_space = model_space
    
    if weighting_space is None:
        weighting_space = model_space
        
    
    #č jsou to informace pro callback 
    estimation={'method': "fast_simplex_estimation", 'nsim':sample_box.nsim}
    estimation['model_space'] = model_space
    estimation['sampling_space'] = sampling_space
    estimation['weighting_space'] = weighting_space
    estimation['outside_budget'] = outside_budget
    estimation['simplex_budget'] = simplex_budget
    estimation['design'] = str(design)
        
    # powerset_correction=True
    #č totíž my zodpovidáme za to, že Triangulaci pošleme správné ISSI 
    issi, outside = _issi_estimate_outside(f, model_space,\
                     sampling_space=sampling_space, powerset_correction=True,\
                design=design, callback=callback, outside_budget=outside_budget)
    
    #č na fastě môžem si dovoliť zkratit cestu 
    failsi = sample_box.failsi
    if np.any(failsi) and not np.all(failsi):
        tri = six.FastSamplingTriangulation(sample_box, tri_space=model_space, issi=issi, \
                 weighting_space=weighting_space, incremental=False,\
                  on_add_simplex=callback, sampling_space=sampling_space, \
                  simplex_budget=simplex_budget, design=design)
                  
        tri.integrate()
                    
        # TRI-compatible estimation
        # -1 = 'outside', 0=success, 1=failure, 2=mix
        estimations = tri.get_pf_estimation()
        #č doplním
        estimations['global_stats']['outside'] = outside
        
        result = {**estimation, **estimations}
        
    elif np.all(failsi):
        # TRI-compatible estimation
        # -1=outside, 0=success, 1=failure, 2=mix
        #č dostaneme vyrovnané odhady Brna-města (-2) a Brna-venkova (-1)
        #č veškerej vnitršek je v poruše
        pf_inside = issi.estimations[-2]
        pf_outside = issi.estimations[-1]
        
        global_stats = dict()
        global_stats['outside'] = outside
        global_stats['success'] = 0
        global_stats['failure'] = 0 # nechce se mi z ISSI vytahovat
        global_stats['mix'] = 0 
        
        # -1=outside, 0=success, 1=failure, 2=mix
        result = {**estimation, \
                'TRI_estimation': {-1:pf_outside, 0:0, 1:pf_inside, 2:0}, \
                'global_stats': global_stats,'vertex_estimation' : pf_inside, \
                'weighted_vertex_estimation' : pf_inside}
        
    else:
        # TRI-compatible estimation
        # -1=outside, 0=success, 1=failure, 2=mix
        #č dostaneme vyrovnané odhady Brna-města (-2) a Brna-venkova (-1)
        #č vnitršek je asi v pořadku
        pf_inside = issi.estimations[-2]
        pf_outside = issi.estimations[-1]
        
        global_stats = dict()
        global_stats['outside'] = outside
        global_stats['success'] = 0 # nechce se mi z ISSI vytahovat
        global_stats['failure'] = 0
        global_stats['mix'] = 0 
        
        # -1=outside, 0=success, 1=failure, 2=mix
        result = {**estimation, \
                'TRI_estimation': {-1:pf_outside, 0:pf_inside, 1:0, 2:0}, \
                'global_stats': global_stats,'vertex_estimation' : 0, \
                'weighted_vertex_estimation' : 0}
            
    # try to put it to the BlackBox
    try:
        sample_box.estimations.append(result)
    except: #č asi ani nebudu neuspěch hlasit
        pass
        
    return result




#č weighting space nechám pro bodové odhady pf
#č ale podle mně nemá význam
def fast_simplex_cubature(sample_box, tn_scheme, model_space='Rn', sampling_space=None, weighting_space=None,\
                                            outside_budget=1000, callback=None, design=None):
    """
    Delaunay triangulation
    """
    
    #č vytahneme ze sample_boxu rozdělení
    f = sample_box.f_model
    
    
    #č zde provadím rozdělení na prostor, ve kterém vzorkujem
    #čs a prostor "modelu", vô ktôrom, v podstatě, měříme vzdaleností
    sampled_plan_model = getattr(sample_box, model_space)
    if sampling_space is None:
        sampling_space = model_space
    
    if weighting_space is None:
        weighting_space = model_space
        
    
    #č jsou to informace pro callback 
    estimation={'method': "fast_simplex_cubature", 'nsim':sample_box.nsim}
    estimation['tn_scheme'] = tn_scheme.name
    estimation['model_space'] = model_space
    estimation['sampling_space'] = sampling_space
    estimation['weighting_space'] = weighting_space
    estimation['outside_budget'] = outside_budget
    estimation['design'] = str(design)
        
    
    # powerset_correction=True
    #č totíž my zodpovidáme za to, že Triangulaci pošleme správné ISSI 
    issi, outside = _issi_estimate_outside(f, model_space,\
                     sampling_space=sampling_space, powerset_correction=True,\
                design=design, callback=callback, outside_budget=outside_budget)
    
    #č na fastě môžem si dovoliť zkratit cestu 
    failsi = sample_box.failsi
    if np.any(failsi) and not np.all(failsi):
        tri = six.FastCubatureTriangulation(sample_box, tn_scheme, tri_space=model_space, issi=issi, \
                 weighting_space=weighting_space, incremental=False,\
                  on_add_simplex=callback)
                  
        tri.integrate()
                    
        # TRI-compatible estimation
        # -1 = 'outside', 0=success, 1=failure, 2=mix
        estimations = tri.get_pf_estimation()
        #č doplním
        estimations['global_stats']['outside'] = outside
        
        result = {**estimation, **estimations}
        
    elif np.all(failsi):
        # TRI-compatible estimation
        # -1=outside, 0=success, 1=failure, 2=mix
        #č dostaneme vyrovnané odhady Brna-města (-2) a Brna-venkova (-1)
        #č veškerej vnitršek je v poruše
        pf_inside = issi.estimations[-2]
        pf_outside = issi.estimations[-1]
        
        global_stats = dict()
        global_stats['outside'] = outside
        global_stats['success'] = 0
        global_stats['failure'] = 0 # nechce se mi z ISSI vytahovat
        global_stats['mix'] = 0 
        
        # -1=outside, 0=success, 1=failure, 2=mix
        result = {**estimation, \
                'TRI_estimation': {-1:pf_outside, 0:0, 1:pf_inside, 2:0}, \
                'global_stats': global_stats,'vertex_estimation' : pf_inside, \
                'weighted_vertex_estimation' : pf_inside}
        
    else:
        # TRI-compatible estimation
        # -1=outside, 0=success, 1=failure, 2=mix
        #č dostaneme vyrovnané odhady Brna-města (-2) a Brna-venkova (-1)
        #č vnitršek je asi v pořadku
        pf_inside = issi.estimations[-2]
        pf_outside = issi.estimations[-1]
        
        global_stats = dict()
        global_stats['outside'] = outside
        global_stats['success'] = 0 # nechce se mi z ISSI vytahovat
        global_stats['failure'] = 0
        global_stats['mix'] = 0 
        
        # -1=outside, 0=success, 1=failure, 2=mix
        result = {**estimation, \
                'TRI_estimation': {-1:pf_outside, 0:pf_inside, 1:0, 2:0}, \
                'global_stats': global_stats,'vertex_estimation' : 0, \
                'weighted_vertex_estimation' : 0}
            
    # try to put it to the BlackBox
    try:
        sample_box.estimations.append(result)
    except: #č asi ani nebudu neuspěch hlasit
        pass
        
    return result
    


#č weighting space nechám pro bodové odhady pf
#č ale podle mně nemá význam
def full_simplex_cubature(sample_box, tn_scheme, model_space='Rn',\
                         weighting_space=None, callback=None):
    """
    Delaunay triangulation
    """
    if weighting_space is None:
        weighting_space = model_space
        
    
    #č jsou to informace pro callback 
    estimation={'method': "full_simplex_cubature", 'nsim':sample_box.nsim}
    estimation['tn_scheme'] = tn_scheme.name
    estimation['model_space'] = model_space
    estimation['weighting_space'] = weighting_space
        
    tri = six.FullCubatureTriangulation(sample_box, tn_scheme, tri_space=model_space, \
             weighting_space=weighting_space, incremental=False,\
              on_add_simplex=callback)
              
    tri.integrate()
                
    # TRI-compatible estimation
    # -1 = 'outside', 0=success, 1=failure, 2=mix
    estimations = tri.get_pf_estimation()
    
    result = {**estimation, **estimations}
        
            
    # try to put it to the BlackBox
    try:
        sample_box.estimations.append(result)
    except: #č asi ani nebudu neuspěch hlasit
        pass
        
    return result



#!/usr/bin/env python
# coding: utf-8

import numpy as np
from scipy.spatial import Delaunay, KDTree
from scipy.optimize import linprog



#
# Qhull
#

class Qframe(Delaunay):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.coplanar.size:
            print("Qframe: coplanar points ", self.coplanar)
            
            
    def is_couple(self, couple_indices):
        i, j = couple_indices
        ii = self.simplices
        return np.any(np.sum((ii == i) | (ii == j), axis=1) == 2)
    
    def generate_wireframe(self, callback=lambda:True):
        n = self.npoints
        size = n * (n - 1) // 2
        self.condensed_contacts = np.zeros(size, dtype=bool)
        
        for simplex in self.simplices:
            self._handle_simplex(simplex)
            if callback():
                break
        
        
    def _handle_simplex(self, simplex):
        if len(simplex) > 3:
            i = simplex[0]
            jj = simplex[1:]
            self._handle_simplex(jj)
            for j in jj:
                self._set_contact(i, j)
        else:
            i, j, k = simplex
            self._set_contact(i, j)
            self._set_contact(j, k)
            self._set_contact(k, i)
            
            
        
    def _set_contact(self, x, y):
        i, j = min(x, y), max(x, y)
        m = self.npoints
        entry = m * i + j - ((i + 2) * (i + 1)) // 2
        self.condensed_contacts[entry] = True




def convex_solver_test(points, tries_to_fix=1, tol=1e-7):
    tri = Delaunay(points)
    CS = ConvexSpline(points)
    
    false_positives = []
    false_negatives = []
    
    ii = tri.simplices
    for i in range(len(points)): 
        for j in range(i):
            tri_status = np.any(np.sum((ii == i) | (ii == j), axis=1) == 2)
            CS_status = CS.is_couple((i,j), tries_to_fix=tries_to_fix, tol=tol)
            if tri_status == CS_status:
                print(i, j, tri_status)
            elif tri_status and not CS_status:
                print(i, j, "False negative! QHull says %s, but CS thinks %s" % (tri_status, CS_status))
                false_negatives.append((i,j))
            else:
                print(i, j, "False positive! QHull says %s, but CS thinks %s" % (tri_status, CS_status))
                false_positives.append((i,j))
    
    if false_positives or false_negatives:
        print("False positives: ", false_positives)
        print("False negatives: ", false_negatives)
    else:
        print("Test succesfully passed")
            




#
# DirectContact
#

#č zda existuje přímy kontakt, prostě na usečce
class DirectContact(KDTree):
    """
    č Obecně, skoro pro každou dimenzi v rozmězí od 2D do 100D
    č nejlevnější bylo 
    č 1. strom s pár bodíků
    č 2. ConvexSolver
    č 3. strom s nsim bodíků
    """
    def is_couple(self, couple_indices, **kwargs):
        i, j = couple_indices
        
        half_point = np.mean(self.data[[i, j]], axis=0)
        
        __dd, ii = self.query(half_point, k=2, **kwargs)
        #č pro jediný bod jednoducha kontrola je rychlejší 
        #č jak volání numpy množinových funkcí
        return (i in ii) and (j in ii)



#
# Gabriel
#

#č zda existuje přímy kontakt, prostě na usečce
class Gabriel:
    def __init__(self, points):
        self.points = points
        nsim, ndim = points.shape
        
        self.b = np.sum(np.square(points), axis=1)
        self._mask = np.empty(nsim, dtype=bool)
        
        
    def is_couple(self, couple_indices):
        i, j = couple_indices
        
        mask = self._mask
        if len(mask) < 3:
            return True
        mask[:] = True
        mask[[i, j]] = False
        
        X = self.points
        b = self.b
        
        test = b[mask] - X[mask] @ X[i] - X[mask] @ X[j] + np.inner(X[i], X[j])
        
        return np.all(test >= 0)
        

#
# ConvexSolver
#


def get_offsets(a, X):
    #č nejak tuším, že v poslední dimenzi 
    #č znaménko normál musí být jednotné
    # a = a * -np.sign(a[-1])
    a = a * (1 - 2 * (a[-1] > 0)) 
    b = X @ a
    return b


#č podoba tamtamtoho tehdejšího prvního pokusu.
#č Jen pro srandu. Pro historickej záznam
def convex_solver(points, line_indices, tries_to_fix=1, tol=1e-7):
    i, j = line_indices
    
    X = points
    __nsim, ndim = X.shape
    
    basis = np.random.random((ndim, ndim))
    
    #č QR rozklad jede po sloupcich
    basis[:, 0] = X[j] - X[i]
    
    for dim in range(1, ndim-1):
        #č co jsem viděl, numpy matici Q normalizuje
        #č a první sloupec zůstavá (skoro) tím samým, co byl před tím
        basis[:, :dim+1], __ = np.linalg.qr(basis[:, :dim+1])
        
        # get constrain
        a = basis[:, dim]
        b = get_offsets(a, X)
        if np.max(b) - np.max(b[[i, j]]) < tol:
            return True
        else:
            idx = np.argmax(b)
            basis[:, dim] = basis[:, 1]
            basis[:, 1] = X[idx] - X[i]
            
            
    
    for __ in range(ndim + tries_to_fix):
        basis, __ = np.linalg.qr(basis)
        
        # get constrain
        a = basis[:, -1]
        b = get_offsets(a, X)
        if np.max(b) - np.max(b[[i, j]]) < tol:
            return True
        else:
            idx = np.argmax(b)
            basis[:, 2:] = basis[:, 1:-1]
            basis[:, 1] = X[idx] - X[i]
    
    basis, __ = np.linalg.qr(basis)
    a = basis[:, -1]
    b = get_offsets(a, X)
    return np.max(b) - np.max(b[[i, j]]) < tol


        
        
#č Hlavní pointa ConvexSpline třídy:
#č Využit navíc geometrických informací, které už předem známé:
#č 1. Známe souřadnice vzorků. 
#č 2. Víme, že přímka mezí těmi dvěma vzorky leží v hyperrovině
#č 3. Vždyť to my zvedáme na povrch convexního paraboloidu!
#č    Můžeme tedy v každém její bodě najit tečnou hyperrovinu.    
def convex_spline(points, couple_indices, tries_to_fix=1, tol=1e-7):
    i, j = couple_indices

    X = points
    __nsim, ndim = X.shape
    
    
    
    
    basis = np.random.random((ndim, ndim))
    #č první vektor musí být zadán přímkou mezí vzorky
    #č jinak se posype náš předpoklad, že leží v hyperrovině
    #č a žádné výsledky "za", "před" hyperrovinou nebudou nic znamenat
    basis[:, 0] = X[j] - X[i] #č QR rozklad jede po sloupcich
    
    #č jako indice zkusme použit bod na usečce uprostřed mezí vzorky
    half_point = np.mean(X[[i, j]], axis=0) 
    
    #č všechy souřadnice jsou dány radius-vektorem od středu,
    #č ale poslední souřadnice je to naše zvednutí, 
    #č kde bychom mohli zkusit dát korrektnější směr.
    #č Mně humpolackými uvahami o tečně paraboly 
    #č na caru paríru vyšlo něco jako
    #č že stačí dát poslední složku 0,5.
    #č Jakože čím je roloměr je větší, 
    #č tím je "svislá" složka automaticky měnší.
    #č Jakože netřeba ani normalizovat, ani nic "složitě" počítat
    half_point[-1] = -0.5
    
    #č ten náš radius-vektor není ortogonální 
    #č k přímce, na které leží ty naši dva vzorky
    #č QR rozklad je ortogonalizuje 
    #č a vygeneruje další ortogonalní vektory
    basis[:, 1] = half_point
    basis, __ = np.linalg.qr(basis)
    
    
    #č vytahneme náš ortogonalizovaný odhad 
    #č normálního vektoru
    a = basis[:, 1]
    b = get_offsets(a, X)
    
    if np.max(b) - np.max(b[[i, j]]) < tol:
        #č bylo to myšleno jako jakési zoufalství
        #č ale funguje překvapivě dobře
        return True 
    else:
        idx = np.argmax(b)
        #č nahradíme v jedničce naš odhad normálního vektoru
        #č nalezenou překažkou.
        #č Zbytek bazí jíž byl ortogonální 
        #č k té naši pomyšlené normále
        basis[:, 1] = X[idx] - X[i]
    
    #č Ve zbytku pokračujeme jako vždycky.
    #č Pořad ndim pokusu u False párečku
    for __ in range(ndim + tries_to_fix):
        basis, __ = np.linalg.qr(basis)
        
        # get constrain
        a = basis[:, -1]
        b = get_offsets(a, X)
        if np.max(b) - np.max(b[[i, j]]) < tol:
            return True
        else:
            idx = np.argmax(b)
            basis[:, 2:] = basis[:, 1:-1]
            basis[:, 1] = X[idx] - X[i]
    
    basis, __ = np.linalg.qr(basis)
    a = basis[:, -1]
    b = get_offsets(a, X)
    return np.max(b) - np.max(b[[i, j]]) < tol







        
#č na rozdil od předchozího spline, 
#č sort nesestavuje hned ortogonální k normále bazi, dělá to postupně
#č klíčovým stalo to, že v každem kroku obnovujeme původní odhad normály
def convex_sort(points, couple_indices, tries_to_fix=1, tol=1e-7):
    i, j = couple_indices

    X = points
    __nsim, ndim = X.shape
    
    
    
    
    basis = np.empty((ndim, ndim))
    #č první vektor musí být zadán přímkou mezí vzorky
    #č jinak se posype náš předpoklad, že leží v hyperrovině
    #č a žádné výsledky "za", "před" hyperrovinou nebudou nic znamenat
    basis[:, 0] = X[j] - X[i] #č QR rozklad jede po sloupcich
    
    #č jako indice zkusme použit bod na usečce uprostřed mezí vzorky
    half_point = np.mean(points[[i, j]], axis=0) 
    
    
    #č všechy souřadnice jsou dány radius-vektorem od středu,
    #č ale poslední souřadnice je to naše zvednutí, 
    #č kde bychom mohli zkusit dát korrektnější směr.
    #č Mně humpolackými uvahami o tečně paraboly 
    #č na caru paríru vyšlo něco jako
    #č že stačí dát poslední složku 0,5.
    #č Jakože čím je roloměr větší, 
    #č tím je "svislá" složka automaticky měnší.
    #č Jakože netřeba ani normalizovat, ani nic "složitě" počítat
    half_point[-1] = -0.5
    basis[:, 1] = half_point
    
    #č ten náš radius-vektor není ortogonální 
    #č k přímce, na které leží ty naši dva vzorky
    #č musíme ho (QR rozkladem) ortogonalizovat
    for dim in range(1, ndim-1):
        #č co jsem viděl, numpy matici Q normalizuje
        #č a první sloupec zůstavá (skoro) tím samým, co byl před tím
        basis[:, :dim+1], __ = np.linalg.qr(basis[:, :dim+1])
        
        # get constrain
        a = basis[:, dim]
        b = get_offsets(a, X)
        if (np.max(b) - np.max(b[[i, j]])) < tol:
            return True
        
        #else:
        idx = np.argmax(b)
        basis[:, 2:dim+1] = basis[:, 1:dim]
        basis[:, 1] = X[idx] - X[i]
        basis[:, dim+1] = half_point
    
    
    
    for __ in range(tries_to_fix):
        basis, __ = np.linalg.qr(basis)
        # get constrain
        a = basis[:, -1]
        b = get_offsets(a, X)
        if (np.max(b) - np.max(b[[i, j]])) < tol:
            return True
        idx = np.argmax(b)
        basis[:, 2:] = basis[:, 1:-1]
        basis[:, 1] = X[idx] - X[i]
    
    
    basis, __ = np.linalg.qr(basis)
    a = basis[:, -1]
    b = get_offsets(a, X)
    return (np.max(b) - np.max(b[[i, j]])) < tol






def convex_sprite(X, couple_indices, tries_to_fix=1, tol=1e-7):
    i, j = couple_indices
    __nsim, ndim = X.shape
    
    # sprite does not mean anything. Just a word
    #č místo vyslovéné bazi budeme vektory ukladat do 
    #č jakési obecné matici
    sprite = np.empty((ndim + tries_to_fix, ndim))
    #č první vektor musí být zadán přímkou mezí vzorky
    #č jinak se posype náš předpoklad, že leží v hyperrovině
    #č a žádné výsledky "za", "před" hyperrovinou nebudou nic znamenat
    baseline = X[j] - X[i]
    sprite[-2] = baseline
    
    #č jako indice zkusme použit bod na usečce uprostřed mezí vzorky
    half_point = np.mean(X[[i, j]], axis=0) 
    
    
    #č všechy souřadnice jsou dány radius-vektorem od středu,
    #č ale poslední souřadnice je to naše zvednutí, 
    #č kde bychom mohli zkusit dát korrektnější směr.
    #č Mně humpolackými uvahami o tečně paraboly 
    #č na caru paríru vyšlo něco jako
    #č že stačí dát poslední složku 0,5.
    #č Jakože čím je roloměr větší, 
    #č tím je "svislá" složka automaticky měnší.
    #č Jakože netřeba ani normalizovat, ani nic "složitě" počítat
    half_point[-1] = -0.5
    sprite[-1] = half_point
    
    #č ten náš radius-vektor není ortogonální 
    #č k přímce, na které leží ty naši dva vzorky
    #č musíme ho (QR rozkladem) ortogonalizovat
    for dim in range(1, ndim-1):
        basis, __ = np.linalg.qr(sprite[-dim-1:].T) 
        
        # get constrain
        a = basis[:, -1]
        b = get_offsets(a, X)
        if (np.max(b) - np.max(b[[i, j]])) < tol:
            return True
        
        #else:
        idx = np.argmax(b)
        sprite[-dim-1] = X[idx] - X[i]
        sprite[-dim-2] = baseline
    
    
    #č tady máme
    #č sprite[-ndim+1] = nějaký vektor
    #č sprite[-ndim] = baseline
    
    for t in range(tries_to_fix):
        basis, __ = np.linalg.qr(sprite[-ndim-t:].T)
        # get constrain
        a = basis[:, -1]
        b = get_offsets(a, X)
        if (np.max(b) - np.max(b[[i, j]])) < tol:
            return True
        idx = np.argmax(b)
        sprite[-ndim-t] = X[idx] - X[i]
        sprite[-ndim-1-t] = baseline
    
    assert (sprite[0] == baseline).all()
    
    basis, __ = np.linalg.qr(sprite.T)
    a = basis[:, -1]
    b = get_offsets(a, X)
    return (np.max(b) - np.max(b[[i, j]])) < tol



def get_longest(O, A, B):
    a_i = O - A
    a_j = O - B
    if np.inner(a_i[:-1], a_i[:-1]) > np.inner(a_j[:-1], a_j[:-1]):
        return a_i
    else:
        return a_j


def convex_slice(X, couple_indices, tries_to_fix=1, tol=1e-7):
    i, j = couple_indices
    nsim, ndim = X.shape
    
    mask = np.full(nsim, True)
    mask[[i, j]] = False
    
    #č jako indice zkusme použit bod na usečce uprostřed mezí vzorky
    #
    #č všechy souřadnice jsou dány radius-vektorem od středu,
    #č ale poslední souřadnice je to naše zvednutí, 
    #č kde bychom mohli zkusit dát korrektnější směr.
    #č Mně humpolackými uvahami o tečně paraboly 
    #č na caru paríru vyšlo něco jako
    #č že stačí dát poslední složku 0,5.
    #č Jakože čím je roloměr větší, 
    #č tím je "svislá" složka automaticky měnší.
    #č Jakože netřeba ani normalizovat, ani nic "složitě" počítat
    half_point = np.mean(X[[i, j]], axis=0) 
    half_point[-1] = -0.5
    
    b = get_offsets(half_point, X)
    if (np.max(b[mask]) - np.min(b[[i, j]])) < tol:
        return True
    
    #č nepovedlo. Tak deme na věc
    idxs = []
    
    # sprite does not mean anything. Just a word
    #č místo vyslovéné bazi budeme vektory ukladat do 
    #č jakési obecné matici
    sprite = np.empty((ndim + tries_to_fix, ndim))
    baseline = X[j] - X[i]
    sprite[-2] = baseline
    sprite[-1] = half_point
    
    
    
    for dim in range(1, ndim-1):
        basis, __ = np.linalg.qr(sprite[-dim-1:].T) 
        
        # get constrain
        a = basis[:, -1]
        b = get_offsets(a, X)
        #č u kontrol musím taky hlídat, aby argmax nebyl i, nebo j
        if (np.max(b[mask]) - np.max(b[[i, j]])) < tol:
            print(idxs)
            return True
        
        #else:
        idx = np.argmax(b)
        idxs.append(idx)
        sprite[-dim-1] = get_longest(X[idx], X[i], X[j])
        sprite[-dim-2] = baseline
    
    
    #č tady máme
    #č sprite[-ndim+1] = nějaký vektor
    #č sprite[-ndim] = baseline
    
    for t in range(tries_to_fix):
        basis, __ = np.linalg.qr(sprite[-ndim-t:].T)
        # get constrain
        a = basis[:, -1]
        b = get_offsets(a, X)
        #č u kontrol musím taky hlídat, aby argmax nebyl i, nebo j
        if (np.max(b[mask]) - np.max(b[[i, j]])) < tol:
            print(idxs)
            return True
        idx = np.argmax(b)
        idxs.append(idx)
        sprite[-ndim-t] = get_longest(X[idx], X[i], X[j])
        sprite[-ndim-1-t] = baseline
    
    assert (sprite[0] == baseline).all()
    
    basis, __ = np.linalg.qr(sprite.T)
    a = basis[:, -1]
    b = get_offsets(a, X)
    print(idxs)
    return (np.max(b[mask]) - np.min(b[[i, j]])) < tol






class ConvexSolver:
    """
    č Hlavní pointa třídy:
    č pokud dva body zvednuté na povrch convexního paraboloidu
    č v prostoru ndim+1 (feature space, quadratic kernel)
    č lze lineárně separovat (hyperrovinou) od ostatních bodů,
    č znamená to, že v původním prostoru 
    č jejich Voroného buňky budou mít společnou stěnu (kontakt),
    č neboli, což je totež, u Delone triangulace by měly společné simplexy
    
    č nebudeme puntičkařit a pro jednoduchost předepíšeme,
    č nechť dva body zájmu leží přímo v hyperrovině
    č (bude to hlasít existence rozhraní i v degenerovaných případech
    č jako např. tří teček na jedné přímce. Mně to ale nevadí)
    """
    def __init__(self, points, convex_solver=convex_sprite):
        nsim, ndim = points.shape
        
        self.lifted_points = np.empty((nsim, ndim + 1))
        self.lifted_points[:, :ndim] = points
        # kind of datascience. feature space, quadratic kernel...
        self.lifted_points[:, -1] = np.sum(np.square(points), axis=1)
        
        self.convex_solver = convex_solver
        
    def is_couple(self, couple_indices, *args, **kwargs):
        return self.convex_solver(self.lifted_points, couple_indices, *args, **kwargs)











class LocalizedHull:
    """
    č Hlavní pointa třídy:
    č pokud dva body zvednuté na povrch convexního paraboloidu
    č v prostoru ndim+1 (feature space, quadratic kernel)
    č lze lineárně separovat (hyperrovinou) od ostatních bodů,
    č znamená to, že v původním prostoru 
    č jejich Voroného buňky budou mít společnou stěnu (kontakt),
    č neboli, což je totež, u Delone triangulace by měly společné simplexy
    
    č nebudeme puntičkařit a pro jednoduchost předepíšeme,
    č nechť dva body zájmu leží přímo v hyperrovině
    č (bude to hlasít existence rozhraní i v degenerovaných případech
    č jako např. tří teček na jedné přímce. Mně to ale nevadí)
    """
    def __init__(self, points):
        nsim, ndim = points.shape
        
        self.points = points
        
        self._basis = np.empty((ndim + 1, ndim + 1), dtype=float)
        self._mask = np.empty(nsim, dtype=bool)
        
        self.lifted_points = np.empty((nsim, ndim + 1))
        self.lifted_points[:, :ndim] = points
        # kind of datascience. feature space, quadratic kernel...
        self.lifted_points[:, -1] = np.sum(np.square(points), axis=1)

    def get_tri_basis(self, tri, idxs):
        for simplex in tri.simplices:
            if (0 in simplex) and (1 in simplex):
                indices = np.array(idxs)[simplex]
                basis = self.lifted_points[indices]
                basis = basis - basis[-1]
                basis, __ = np.linalg.qr(basis.T)
                return True, basis
        return False, None
                
            
        
    def is_couple(self, couple_indices, tol=1e-7):
        i, j = couple_indices
        X = self.lifted_points
        nsim, ndim = X.shape
        
        if nsim < 3:
            return True
        
        mask = self._mask
        mask[:] = True
        mask[[i, j]] = False
        
        #č jako indice zkusme použit bod na usečce uprostřed mezí vzorky
        #
        #č všechy souřadnice jsou dány radius-vektorem od středu,
        #č ale poslední souřadnice je to naše zvednutí, 
        #č kde bychom mohli zkusit dát korrektnější směr.
        #č Mně humpolackými uvahami o tečně paraboly 
        #č na caru paríru vyšlo něco jako
        #č že stačí dát poslední složku 0,5.
        #č Jakože čím je roloměr větší, 
        #č tím je "svislá" složka automaticky měnší.
        #č Jakože netřeba ani normalizovat, ani nic "složitě" počítat
        half_point = np.mean(X[[i, j]], axis=0) 
        half_point[-1] = -0.5
        
        b = get_offsets(half_point, X)
        if (np.max(b[mask]) - np.min(b[[i, j]])) < tol:
            return True
        
        #č nepovedlo. Tak deme na věc
        idxs = [i, j]
        basis = self._basis
        #č první vektor musí být zadán přímkou mezí vzorky
        #č jinak se posype náš předpoklad, že leží v hyperrovině
        #č a žádné výsledky "za", "před" hyperrovinou nebudou nic znamenat
        basis[:, 0] = X[j] - X[i] #č QR rozklad jede po sloupcich
        basis[:, 1] = half_point
        
        #č ten náš radius-vektor není ortogonální 
        #č k přímce, na které leží ty naši dva vzorky
        #č musíme ho (QR rozkladem) ortogonalizovat
        for dim in range(1, ndim-1):
            #č co jsem viděl, numpy matici Q normalizuje
            #č a první sloupec zůstavá (skoro) tím samým, co byl před tím
            basis[:, :dim+1], __ = np.linalg.qr(basis[:, :dim+1])
            
            # get constrain
            a = basis[:, dim]
            b = get_offsets(a, X)
            if (np.max(b) - np.max(b[[i, j]])) < tol:
                return True
            
            #else:
            idx = np.argmax(b)
            idxs.append(idx)
            basis[:, 2:dim+1] = basis[:, 1:dim]
            basis[:, 1] = X[idx] - X[i]
            basis[:, dim+1] = half_point
        
        #č Doufám, že idxs budou unikatní
        assert len(idxs) == len(set(idxs))
        
        #č v tuhle chvili máme ndim indexů
        #č což bude stačít na jeden simplex 
        #č v původním ndim-1 prostoru.
        indices = np.array(idxs)
        basis = X[indices]
        basis = basis - basis[-1]
        basis, __ = np.linalg.qr(basis.T)
        # get constrain
        a = basis[:, -1]
        b = get_offsets(a, X)
        if (np.max(b) - np.max(b[[i, j]])) < tol:
            return True
        
        #č Doufám, že idxs budou unikatní
        assert len(idxs) == len(set(idxs))
        
        idx = np.argmax(b)
        idxs.append(idx)
        
        #č Dále už jedeme triangulací
        tri = Delaunay(self.points[idxs], incremental=True)
        
        while True:
            if tri.coplanar.size:
                return True
                
            status, basis = self.get_tri_basis(tri, idxs)
            if not status:
                return False
            
            # get constrain
            a = basis[:, -1]
            b = get_offsets(a, X)
            if (np.max(b) - np.max(b[[i, j]])) < tol:
                return True
            
            #č Doufám, že idxs budou unikatní
            assert len(idxs) == len(set(idxs))
            
            idx = np.argmax(b)
            idxs.append(idx)
            tri.add_points(np.atleast_2d(self.points[idx]))





#
# Voronoi adjacency by linear programming
#



class LinearSolver:
    """
    linprog:
    minimize c @ x
    such that:
    A_ub @ x <= b_ub
    A_eq @ x == b_eq
    lb <= x <= ub
    """
    
    def __init__(self, points):
        self.gabriel = Gabriel(points)
        
        
        self.points = points
        nsim, ndim = points.shape
        size = nsim * (nsim - 1) // 2
        self.condensed_contacts = np.zeros(size, dtype=bool)
        self.arange = np.arange(nsim)
        self.nsim = nsim
        
        self.A = A = np.empty((ndim + 1, nsim))
        A[:-1] = points.T
        A[-1] = 1
        
        self._b = np.empty(ndim + 1)
        self._b[-1] = 1
        
        self.c = np.sum(np.square(points), axis=1)
        
    def LP(self, couple_indices):
        i, j = couple_indices
        X = self.points
        
        midpoint = np.mean(X[[i, j]], axis=0)
        b_eq = self._b
        b_eq[:-1] = midpoint
        
        return linprog(self.c, A_eq=self.A, b_eq=b_eq)
        
    def is_couple(self, couple_indices, tol=1e-5):
        i, j = couple_indices
        if self._get_contact(i, j):
            return True
            
        if self.gabriel.is_couple(couple_indices):
            self._set_contact(i, j)
            return True
        
        result = self.LP(couple_indices)
        mask = result.x > tol
        self._handle_polytope(self.arange[mask])
        
        return mask[i] and mask[j]
        
        #X = self.points
        #test = (np.inner(X[i], X[i]) + np.inner(X[j], X[j])) / 2
        #return test <= (result.fun + tol)
        
    
    
    def _handle_polytope(self, polytope):
        if len(polytope) > 1:
            i = polytope[0]
            jj = polytope[1:]
            self._handle_polytope(jj)
            for j in jj:
                self._set_contact(i, j)
        
    def _set_contact(self, x, y):
        i, j = min(x, y), max(x, y)
        entry = self.nsim * i + j - ((i + 2) * (i + 1)) // 2
        self.condensed_contacts[entry] = True
        
    def _get_contact(self, x, y):
        i, j = min(x, y), max(x, y)
        entry = self.nsim * i + j - ((i + 2) * (i + 1)) // 2
        return self.condensed_contacts[entry]
        



class ConvexLinear(LinearSolver):
    def __init__(self, points, convex_solver=convex_sprite):
        super().__init__(points)
        
        nsim, ndim = points.shape
        self.lifted_points = np.empty((nsim, ndim + 1))
        self.lifted_points[:, :ndim] = points
        # kind of datascience. feature space, quadratic kernel...
        self.lifted_points[:, -1] = np.sum(np.square(points), axis=1)
        
        self.convex_solver = convex_solver
    
    def is_couple(self, couple_indices, tol=1e-5, **kwargs):
        i, j = couple_indices
        if self._get_contact(i, j):
            return True
            
        if self.gabriel.is_couple(couple_indices):
            self._set_contact(i, j)
            return True
        
        if self.convex_solver(self.lifted_points, couple_indices,  tol=tol, **kwargs):
            self._set_contact(i, j)
            return True
        
        result = self.LP(couple_indices)
        mask = result.x > tol
        self._handle_polytope(self.arange[mask])
        
        return mask[i] and mask[j]
    





#!/usr/bin/env python
# coding: utf-8

import numpy as np
from . import sball
from scipy import stats
from .candynodes import CandyNodes
from .IS_stat import get_1DS_sample # for Shell_1DS

from collections import namedtuple

try: # try to outthink OS's memory management
    import os
    mem_bytes = os.sysconf('SC_PAGE_SIZE') * os.sysconf('SC_PHYS_PAGES')
except BaseException as e:
    mem_GB = 16
    mem_bytes = mem_GB * 1024**3 # hello, Windows!
    #print("ghull failed to get amount of RAM installed. %s GB is supposed."% mem_GB, repr(e))
    #print("BTW, you are using Windows, aren't you?")
    
    
    
ShellStats = namedtuple('ShellStats', (
                                        'nsim',
                                        'nvar',
                                        'nfacets',
                                        'r', 'R',
                                        'inner',
                                        'shell',
                                        'outer',
                                        'FORM_outside',
                                        'TwoFORM_outside',
                                        'orth_outside'
                                        ))

ShellEstimation = namedtuple('ShellEstimation', 
                            ('shell_budget', 'shell_inside', 
                            'shell_outside', 'inside', 'outside'))


#č pomocná třída pro integrování
#č nejsem tedy jist, jestli je to nejlepší napad - 
#č dělit Ghull na podtřídy, delegovat funkcionalitu
#č ale jinak Ghull se stavá hodně překomplikovaným.
#č nic lepšího mně nenapadá, přemyšlel jsem dlouho.
class Shell_1DS:
    """1DS stands for 1D sampling.
    1DS is, actually, an importance sampling method
    that, actually, does not calculate IS weights
    as f_density / h_density, but
    (arbitrary, nonuniformly) divides interval to subintervals.
    By using CDF transformes subintervals to 0-1 measure line.
    Then gets one node as middle point of every subinterval,
    weights therefore are just interval widths itself.
    No sampling imprecisions are introduced, 
    therefore no spring, no correction are needed. 
    """
    def __init__(self, hull, shell):
        self.shell = shell
        self.hull = hull
        self.nvar = hull.sample.nvar
        
        #č objevilo se, že scipy.stats.chi je neunosně nepřesné
        #č vzdává se jíž na poloměru 8
        self.sball = sball.Sball(self.nvar)
        
        self.integration_cutoff = np.inf
        
    
    def integrate(self, nis, callback_all=None, callback_outside=None):
        self.reset(nis)
        
        if self.hull.nsimplex == 0:
            bus = self.integration_cutoff
        else:
            bus = int(mem_bytes / self.hull.nsimplex / 8 / 10) + 1
        while self.nsampled < nis:
            
            seats = min(nis - self.nsampled, self.integration_cutoff, bus)
            try: 
                self._process_part(seats, nis, callback_all, callback_outside)
            except MemoryError as m:
                print(self.__class__.__name__ +":", "memory error, %s sedaček" % seats, repr(m))
                self.integration_cutoff = int(np.ceil(seats/2))
        
        assert nis == self.nsampled
        
        return self._get_result()
        
    
    def reset(self, nis): # clear
        
        self.nsampled = 0
        
        #č poloměry bereme ze skořapky
        #č za správné nastavení (stejně jako u MC)
        #č zodpovidá uživatel třídy
        #č třída vlastní odhad r nijak nevyuživá!
        r = self.shell.r
        R = self.shell.R
        # let's predefine 1D sequence at very beginning
        if r > np.sqrt(self.nvar - 1):
            x_sub = np.geomspace(r, R, nis+1, endpoint=True)
        else:
            x_sub = np.linspace(r, R, nis+1, endpoint=True)
        
        
        x, weights = get_1DS_sample(self.sball, x_sub)
        self.x = x
        self.weights = weights
        self.mask = np.empty(nis, dtype=bool)
        
    
    
    # bus analogy
    def _process_part(self, seats, nis, callback_all=None, callback_outside=None):
        # boarding
        left = self.nsampled
        right = self.nsampled + seats
        rs = self.x[left:right]
        
        # rand_dir: prepare ns random directions on a unit d-sphere
        rand_dir = sball.get_random_directions(seats, self.nvar) #random directions
        nodes_G = rand_dir*rs[:,None]
        nodes = self.hull.sample.f_model.new_sample(nodes_G, space='G')
        
        # who is back?
        d, i = self.hull.query(nodes)
        mask = d > 0
        assert len(mask) == seats
        
        # the most important part
        self.nsampled += len(mask) 
        self.mask[left:right] = mask
        
        if callback_all is not None:
            # -2 = 'inside' -1 = 'outside'
            candy_nodes = CandyNodes(nodes, event_id=mask-2, is_outside=mask, d=d, i=i)
            callback_all(candy_nodes)
            
        if (callback_outside is not None) and np.any(mask):
            callback_outside(nodes[mask], d[mask], i[mask])
        
        
        
    def _get_result(self):
        #č mask related to hull.is_outside()
        shell_pf = np.sum(self.weights[self.mask])
        shell_ps = np.sum(self.weights[~self.mask])
        
        return self.nsampled, shell_ps, shell_pf
        
        
        
        
class GaussianAnnulus:
    def __init__(self, hull):
        assert hull.space == 'G'
        
        self.hull = hull
        self.shell = sball.Shell(hull.sample.nvar)
        self.outside_dist = sball.Shell(hull.sample.nvar)
        self.sample = hull.sample
        
        self.gint = Shell_1DS(self.hull, self.shell)
        
            
    def boom(self, ns, use_MC=False):
        if use_MC:
            self.outside_dist.set_bounds(self.get_R())
            nodes_G = self.outside_dist.rvs(ns)
        else:
            # rand_dir: prepare ns random directions on a unit d-sphere
            rand_dir = sball.get_random_directions(ns, self.sample.nvar) #random directions
            
            #č deme od vnější .get_R() kružnici směrem ven
            r = self.get_R()
            
            # maximum radius, where norm.pdf() wasn't zero
            # -38.575500173381374935388521407730877399444580
            # don't ask me what the magic python use to distinguish
            # digits after double precision
            max_R_ever = 37
            if r < max_R_ever:
                R = max_R_ever
            else:
                R = r + 10
            r = np.linspace(self.get_R(), max_R_ever, ns, endpoint=True) 
            nodes_G = rand_dir*r[:,None]
        
        nodes = self.sample.f_model.new_sample(nodes_G, space='G')
        return nodes
        
            
    def get_R(self):
        sum_squared = np.sum(np.square(self.sample.G), axis=1)
        #index = np.argmax(sum_squared)
        return np.sqrt(np.nanmax(sum_squared))
        
    def setup_shell(self):
        shell = self.shell
        r = self.hull.get_r()
        R = self.get_R()
        
        if r<0:
            shell.set_bounds(0, R)
        else:
            shell.set_bounds(r, R)
        
        return r, R
    
    def get_shell_estimation(self):
        r, R = self.setup_shell()
        
        return ShellStats(
                            self.sample.nsim,
                            self.sample.nvar,
                            self.hull.nsimplex,
                            r, R,
                            self.shell.ps,
                            self.shell.p_shell,
                            self.shell.pf,
                            stats.norm.sf(r),
                            self.hull.get_2FORM_outside(),
                            self.hull.get_orth_outside()
                        )
        
    def integrate(self, nis, callback_all=None, callback_outside=None):
        self.setup_shell()
        
        nsampled, shell_ps, shell_pf = self.gint.integrate(nis, callback_all, callback_outside)
        
        inside = shell_ps + self.shell.ps
        outside = shell_pf + self.shell.pf
        
        return ShellEstimation(nsampled, shell_ps, shell_pf, inside, outside)
        
        
